import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component


@cell
def test_astar_pass() -> Component:
    c = gf.Component("get_route_astar")
    w = gf.components.straight()
    left = c << w
    right = c << w
    right.move((100, 80))

    obstacle = gf.components.rectangle(size=(100, 10))
    obstacle1 = c << obstacle
    obstacle2 = c << obstacle
    obstacle1.ymin = 40
    obstacle2.xmin = 25

    port1 = left.ports["o2"]
    port2 = right.ports["o2"]

    route = gf.routing.get_route_astar(
        component=c,
        port1=port1,
        port2=port2,
        resolution=5,
        distance=5.5,
        radius=5,
    )
    c.add(route.references)
    route_length = 179.908
    assert route.length == route_length, print(f"route_length = {route.length}")
    return c


# @cell
# def test_astar_fail() -> Component:
#     c = gf.Component()
#     rect1 = c << gf.components.rectangle()
#     rect2 = c << gf.components.rectangle()
#     rect3 = c << gf.components.rectangle((2, 2))
#     rect2.move(destination=(8, 4))
#     rect3.move(destination=(5.5, 1.5))

#     port1 = Port(
#         "o1", 0, rect1.center + (0, 3), cross_section=gf.get_cross_section("strip")
#     )
#     port2 = port1.copy("o2")
#     port2.orientation = 180
#     port2.center = rect2.center + (0, -3)
#     c.add_ports([port1, port2])

#     route = get_route_astar(c, port1, port2, radius=0.9, width=0.5)
#     c.add(route.references)
#     return c


if __name__ == "__main__":
    c = test_astar_pass()
    # c = test_astar_fail()
    c.show(show_ports=True)



from typing import Tuple, Union

import gdspy
import numpy as np
from gdspy import clipper

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.component_layout import Polygon, _parse_layer
from gdsfactory.component_reference import ComponentReference
from gdsfactory.geometry.offset import _crop_edge_polygons, _polygons_to_bboxes
from gdsfactory.types import ComponentOrReference, Int2, LayerSpec


def _boolean_region(
    all_polygons_A,
    all_polygons_B,
    bboxes_A,
    bboxes_B,
    left,
    bottom,
    right,
    top,
    operation="and",
    precision: float = 1e-4,
):
    """Returns boolean for a region.

    Taking a region of e.g. size (x, y) which needs to be booleaned,
    this function crops out a region (x, y) large from each set of polygons
    (A and B), booleans that cropped region and returns the result.

    Args:
        all_polygons_A : PolygonSet or list of polygons
            Set or list of polygons to be booleaned.
        all_polygons_B : PolygonSet or list of polygons
            Set or list of polygons to be booleaned.
        bboxes_A : list
            List of all polygon bboxes in all_polygons_A
        bboxes_B : list
            List of all polygon bboxes in all_polygons_B
        left : int or float
            The x-coordinate of the lefthand boundary.
        bottom : int or float
            The y-coordinate of the bottom boundary.
        right : int or float
            The x-coordinate of the righthand boundary.
        top : int or float
            The y-coordinate of the top boundary.
        operation : {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}
            Boolean operation to perform.
        precision : float
            Desired precision for rounding vertex coordinates.

    Returns:
        polygons_boolean : PolygonSet or list of polygons
        Set or list of polygons with boolean operation applied.
    """
    polygons_to_boolean_A = _crop_edge_polygons(
        all_polygons_A, bboxes_A, left, bottom, right, top, precision
    )
    polygons_to_boolean_B = _crop_edge_polygons(
        all_polygons_B, bboxes_B, left, bottom, right, top, precision
    )
    return clipper.clip(
        polygons_to_boolean_A, polygons_to_boolean_B, operation, 1 / precision
    )


def _boolean_polygons_parallel(
    polygons_A, polygons_B, num_divisions=(10, 10), operation="and", precision=1e-4
):
    """Returns boolean on a list of subsections of the original geometry.

    Returns list of polygons, with all the booleaned polygons from each subsection.

    Args:
        polygons_A : PolygonSet or list of polygons
            Set or list of polygons to be booleaned.
        polygons_B : PolygonSet or list of polygons
            Set or list of polygons to be booleaned.
        num_divisions : array-like[2] of int
            The number of divisions with which the geometry is divided into
            multiple rectangular regions. This allows for each region to be
            processed sequentially, which is more computationally efficient.
        operation : {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}
            Boolean operation to perform.
        precision : float
            Desired precision for rounding vertex coordinates.

    """
    # Build bounding boxes
    polygons_A = np.asarray(polygons_A)
    polygons_B = np.asarray(polygons_B)
    bboxes_A = _polygons_to_bboxes(polygons_A)
    bboxes_B = _polygons_to_bboxes(polygons_B)

    xmin, ymin = np.min(
        [np.min(bboxes_A[:, 0:2], axis=0), np.min(bboxes_B[:, 0:2], axis=0)], axis=0
    )
    xmax, ymax = np.max(
        [np.max(bboxes_A[:, 2:4], axis=0), np.max(bboxes_B[:, 2:4], axis=0)], axis=0
    )

    xsize = xmax - xmin
    ysize = ymax - ymin
    xdelta = xsize / num_divisions[0]
    ydelta = ysize / num_divisions[1]
    xcorners = xmin + np.arange(num_divisions[0]) * xdelta
    ycorners = ymin + np.arange(num_divisions[1]) * ydelta

    boolean_polygons = []
    for xc in xcorners:
        for yc in ycorners:
            left = xc
            right = xc + xdelta
            bottom = yc
            top = yc + ydelta
            _boolean_region_polygons = _boolean_region(
                polygons_A,
                polygons_B,
                bboxes_A,
                bboxes_B,
                left,
                bottom,
                right,
                top,
                operation=operation,
                precision=precision,
            )
            boolean_polygons += _boolean_region_polygons

    return boolean_polygons


@gf.cell
def boolean(
    A: Union[ComponentOrReference, Tuple[ComponentOrReference, ...]],
    B: Union[ComponentOrReference, Tuple[ComponentOrReference, ...]],
    operation: str,
    precision: float = 1e-4,
    num_divisions: Union[int, Int2] = (1, 1),
    max_points: int = 4000,
    layer: LayerSpec = (1, 0),
) -> Component:
    """Performs boolean operations between 2 Component/Reference/list objects.

    ``operation`` should be one of {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
    Note that 'A+B' is equivalent to 'or', 'A-B' is equivalent to 'not', and
    'B-A' is equivalent to 'not' with the operands switched

    based on phidl.geometry.boolean
    You can also use gdsfactory.drc.boolean_klayout

    Args:
        A: Component(/Reference) or list of Component(/References).
        B: Component(/Reference) or list of Component(/References).
        operation: {'not', 'and', 'or', 'xor', 'A-B', 'B-A', 'A+B'}.
        precision: float Desired precision for rounding vertex coordinates..
        num_divisions: number of divisions with which the geometry is divided into
          multiple rectangular regions. This allows for each region to be
          processed sequentially, which is more computationally efficient.
        max_points: The maximum number of vertices within the resulting polygon.
        layer: Specific layer to put polygon geometry on.

    Returns: Component with polygon(s) of the boolean operations between
      the 2 input Components performed.

    Notes
    -----
    'A+B' is equivalent to 'or'.
    'A-B' is equivalent to 'not'.
    'B-A' is equivalent to 'not' with the operands switched.

    """
    D = Component()
    A_polys = []
    B_polys = []
    A = list(A) if isinstance(A, (list, tuple)) else [A]
    B = list(B) if isinstance(B, (list, tuple)) else [B]

    for X, polys in ((A, A_polys), (B, B_polys)):
        for e in X:
            if isinstance(e, (Component, ComponentReference)):
                polys.extend(e.get_polygons())
            elif isinstance(e, Polygon):
                polys.extend(e.polygons)

    layer = gf.pdk.get_layer(layer)
    gds_layer, gds_datatype = _parse_layer(layer)

    operation = operation.lower().replace(" ", "")
    if operation == "a-b":
        operation = "not"
    elif operation == "b-a":
        operation = "not"
        A_polys, B_polys = B_polys, A_polys
    elif operation == "a+b":
        operation = "or"
    elif operation not in ["not", "and", "or", "xor", "a-b", "b-a", "a+b"]:
        raise ValueError(
            "gdsfactory.geometry.boolean() `operation` "
            "parameter not recognized, must be one of the "
            "following:  'not', 'and', 'or', 'xor', 'A-B', "
            "'B-A', 'A+B'"
        )

    # Check for trivial solutions
    if (not A_polys or not B_polys) and operation != "or":
        if (
            operation != "not"
            and operation != "and"
            and operation == "xor"
            and not A_polys
            and not B_polys
            or operation != "not"
            and operation == "and"
        ):
            p = None
        elif operation != "not" and operation == "xor" and not A_polys:
            p = B_polys
        elif operation != "not" and operation == "xor":
            p = A_polys
        elif operation == "not":
            p = A_polys or None
    elif not A_polys and not B_polys:
        p = None
    elif all(np.array(num_divisions) == np.array([1, 1])):
        p = gdspy.boolean(
            operand1=A_polys,
            operand2=B_polys,
            operation=operation,
            precision=precision,
            max_points=max_points,
            layer=gds_layer,
            datatype=gds_datatype,
        )
    else:
        p = _boolean_polygons_parallel(
            polygons_A=A_polys,
            polygons_B=B_polys,
            num_divisions=num_divisions,
            operation=operation,
            precision=precision,
        )

    if p is not None:
        polygons = D.add_polygon(p, layer=layer)
        [
            polygon.fracture(max_points=max_points, precision=precision)
            for polygon in polygons
        ]
    return D


def test_boolean() -> None:
    c = gf.Component()
    e1 = c << gf.components.ellipse()
    e2 = c << gf.components.ellipse(radii=(10, 6))
    e3 = c << gf.components.ellipse(radii=(10, 4))
    e3.movex(5)
    e2.movex(2)
    c = boolean(A=[e1, e3], B=e2, operation="A-B")
    assert len(c.polygons) == 2, len(c.polygons)


if __name__ == "__main__":
    c = gf.Component()
    e1 = c << gf.components.ellipse()
    e2 = c << gf.components.ellipse(radii=(10, 6))
    e3 = c << gf.components.ellipse(radii=(10, 4))
    e3.movex(5)
    e2.movex(2)
    c = boolean(A=[e1, e3], B=e2, operation="A-B")
    c.show(show_ports=True)



from typing import Tuple

import gdsfactory as gf
from gdsfactory.types import LayerSpec


@gf.cell
def litho_ruler(
    height: float = 2,
    width: float = 0.5,
    spacing: float = 2.0,
    scale: Tuple[float, ...] = (3, 1, 1, 1, 1, 2, 1, 1, 1, 1),
    num_marks: int = 21,
    layer: LayerSpec = "WG",
) -> gf.Component:
    """Ruler structure for lithographic measurement.

    Includes marks of varying scales to allow for easy reading by eye.

    based on phidl.geometry

    Args:
        height: Height of the ruling marks in um.
        width: Width of the ruling marks in um.
        spacing: Center-to-center spacing of the ruling marks in um.
        scale: Height scale pattern of marks.
        num_marks: Total number of marks to generate.
        layer: Specific layer to put the ruler geometry on.
    """
    D = gf.Component()
    for n in range(num_marks):
        h = height * scale[n % len(scale)]
        D << gf.components.rectangle(size=(width, h), layer=layer)

    D.distribute(direction="x", spacing=spacing, separation=False, edge="x")
    D.align(alignment="ymin")
    D.flatten()
    return D


if __name__ == "__main__":
    c = litho_ruler()
    c.show(show_ports=True)



"""Compute modes of a rectangular Si strip waveguide on top of oxide. Note that \
you should only pay attention, here, to the guided modes, which are the modes \
whose frequency falls under the light line -- that is, frequency < beta / 1.45, \
where 1.45 is the SiO2 index.

Since there's no special lengthscale here, you can just
use microns. In general, if you use units of x, the frequencies
output are equivalent to x/lambda# so here, the frequencies will be
output as um/lambda, e.g. 1.5um would correspond to the frequency
1/1.5 = 0.6667.

"""
import pathlib
import pickle
from typing import Dict, Optional

import meep as mp
import numpy as np
from meep import mpb

from gdsfactory.config import CONFIG
from gdsfactory.simulation.disable_print import disable_print, enable_print
from gdsfactory.simulation.get_sparameters_path import get_kwargs_hash
from gdsfactory.simulation.modes.get_mode_solver_cross_section import (
    get_mode_solver_cross_section,
)
from gdsfactory.simulation.modes.types import Mode
from gdsfactory.types import CrossSectionSpec, PathType

mpb.Verbosity(0)


def find_modes_cross_section(
    cross_section: CrossSectionSpec,
    tol: float = 1e-6,
    wavelength: float = 1.55,
    mode_number: int = 1,
    parity=mp.NO_PARITY,
    cache: Optional[PathType] = CONFIG["modes"],
    overwrite: bool = False,
    **kwargs,
) -> Dict[int, Mode]:
    """Computes mode effective and group index for a CrossSection.

    Args:
        cross_section: CrossSection to solve
        tol: tolerance when finding modes.
        wavelength: wavelength in um.
        mode_number: mode order of the first mode.
        parity: mp.ODD_Y mp.EVEN_X for TE, mp.EVEN_Y for TM.
        cache: directory path to cache modes. None disables the file cache.
        overwrite: forces simulating again.
        kwargs: waveguide settings.

    Keyword Args:
        cross_section: CrossSection to simulate
        sy: simulation region width (um).
        sz: simulation region height (um).
        resolution: resolution (pixels/um).
        nmodes: number of modes to compute.
        wavelength: wavelength at which to compute mode (um)
        clad_mat: cladding material around CrossSection
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        dispersive: use dispersive material models (requires higher resolution).
        material_name_to_meep: dispersive materials have a wavelength
            dependent index. Maps layer_stack names with meep material database names.

    Returns: Dict[mode_number, Mode]

    compute mode_number of the lowest frequencies as a function of k. Also display
    "parities", i.e. whether the mode is symmetric or anti_symmetric
    through the y=0 and z=0 planes.
    mode_solver.run(mpb.display_yparities, mpb.display_zparities)

    Above, we outputted the dispersion relation: frequency (omega) as a
    function of wavevector kx (beta). Alternatively, you can compute
    beta for a given omega -- for example, you might want to find the
    modes and wavevectors at a fixed wavelength of 1.55 microns. You
    can do that using the find_k function:

    """
    modes = {}
    mode_solver = get_mode_solver_cross_section(
        cross_section=cross_section, wavelength=wavelength, **kwargs
    )

    nmodes = mode_solver.nmodes
    omega = 1 / wavelength

    h = get_kwargs_hash(
        wavelength=wavelength,
        parity=parity,
        **kwargs,
    )

    if cache:
        cache = pathlib.Path(cache)
        cache.mkdir(exist_ok=True, parents=True)
        filepath = cache / f"{h}_{mode_number}.pkl"

        if filepath.exists() and not overwrite:
            for index, i in enumerate(range(mode_number, mode_number + nmodes)):
                filepath = cache / f"{h}_{index}.pkl"
                mode = pickle.loads(filepath.read_bytes())
                modes[i] = mode
            return modes

    # Output the x component of the Poynting vector for mode_number bands at omega
    disable_print()
    k = mode_solver.find_k(
        parity,
        omega,
        mode_number,
        mode_number + nmodes,
        mp.Vector3(1),
        tol,
        omega * 2.02,
        omega * 0.01,
        omega * 10,
        # mpb.output_poynting_x,
        mpb.display_yparities,
        mpb.display_group_velocities,
    )
    enable_print()
    neff = np.array(k) * wavelength

    # vg = mode_solver.compute_group_velocities()
    # vg = vg[0]
    # ng = 1 / np.array(vg)

    for index, i in enumerate(range(mode_number, mode_number + nmodes)):
        Ei = mode_solver.get_efield(i)
        Hi = mode_solver.get_hfield(i)
        y_num = np.shape(Ei[:, :, 0, 0])[0]
        z_num = np.shape(Ei[:, :, 0, 0])[1]

        modes[i] = Mode(
            mode_number=i,
            neff=neff[index],
            wavelength=wavelength,
            E=Ei,
            H=Hi,
            eps=mode_solver.get_epsilon().T,
            y=np.linspace(
                -1 * mode_solver.info["sy"] / 2.0,
                mode_solver.info["sy"] / 2.0,
                y_num,
            ),
            z=np.linspace(
                -1 * mode_solver.info["sz"] / 2.0,
                mode_solver.info["sz"] / 2.0,
                z_num,
            ),
            material_indices=mode_solver.info["material_indices"],
        )
        if cache:
            filepath = cache / f"{h}_{index}.pkl"
            filepath.write_bytes(pickle.dumps(modes[i]))

    return modes


if __name__ == "__main__":
    import gdsfactory as gf

    cross_section = "strip"
    s = gf.components.straight(cross_section=cross_section)

    s.show()

    m = find_modes_cross_section(
        cross_section=cross_section,
    )

    print(m)

    m1 = m[1]

    print(f"Mode 1 n_eff: {m1.neff}")

    m1.plot_e_all()
    # print(f'Mode 1 n_g: {m1.ng}')

    # print(np.shape(m1.y))
    # print(np.shape(m1.z))
    # print(np.shape(m1.E[:, :, 0, 1]))

    # tol: float = 1e-6
    # wavelength: float = 1.55
    # mode_number: int = 1
    # parity = mp.NO_PARITY
    # mode_solver = get_mode_solver_rib()
    # nmodes = mode_solver.nmodes
    # omega = 1 / wavelength

    # # Output the x component of the Poynting vector for mode_number bands at omega
    # k = mode_solver.find_k(
    #     parity,
    #     omega,
    #     mode_number,
    #     mode_number + nmodes,
    #     mp.Vector3(1),
    #     tol,
    #     omega * 2.02,
    #     omega * 0.01,
    #     omega * 10,
    #     mpb.output_poynting_x,
    #     mpb.display_yparities,
    #     mpb.display_group_velocities,
    # )
    # enable_print()
    # vg = mode_solver.compute_group_velocities()
    # vg0 = vg[0]
    # neff = np.array(k) * wavelength
    # ng = 1 / np.array(vg0)

    # modes = {
    #     i: Mode(
    #         mode_number=i,
    #         neff=neff[index],
    #         solver=mode_solver,
    #         wavelength=wavelength,
    #     )
    #     for index, i in enumerate(range(mode_number, mode_number + nmodes))
    # }



import gdsfactory as gf


def test_path_transition_class():
    P = gf.path.straight(length=10, npoints=101)

    s = gf.Section(width=3, offset=0, layer=gf.LAYER.SLAB90)
    X1 = gf.CrossSection(
        width=1,
        offset=0,
        layer=gf.LAYER.WG,
        name="core",
        port_names=("o1", "o2"),
        sections=[s],
    )

    X2 = gf.CrossSection(
        width=3, offset=0, layer=gf.LAYER.WG, name="core", port_names=("o1", "o2")
    )

    T = gf.path.transition(X1, X2)
    return gf.path.extrude(P, T)


def test_path_transition_function():
    P = gf.path.straight(length=10, npoints=101)
    X1 = gf.cross_section.cross_section(width=1)
    X2 = gf.cross_section.cross_section(width=3)
    T = gf.path.transition(X1, X2)
    P = gf.path.straight(length=10, npoints=101)
    return gf.path.extrude(P, T)


def test_path_transitions():
    import gdsfactory as gf

    # Create our first CrossSection
    s1 = gf.Section(width=2.2, offset=0, layer=(3, 0), name="etch")
    s2 = gf.Section(width=1.1, offset=3, layer=(1, 0), name="wg2")
    X1 = gf.CrossSection(
        width=1.2,
        offset=0,
        layer=(2, 0),
        name="wg",
        port_names=("o1", "o2"),
        sections=[s1, s2],
    )

    # Create the second CrossSection that we want to transition to
    s1 = gf.Section(width=3.5, offset=0, layer=(3, 0), name="etch")
    s2 = gf.Section(width=3, offset=5, layer=(1, 0), name="wg2")
    X2 = gf.CrossSection(
        width=1,
        offset=0,
        layer=(2, 0),
        name="wg",
        port_names=("o1", "o2"),
        sections=[s1, s2],
    )

    # To show the cross-sections, let's create two Paths and
    # create Components by extruding them
    P1 = gf.path.straight(length=5)
    P2 = gf.path.straight(length=5)
    wg1 = gf.path.extrude(P1, X1)
    wg2 = gf.path.extrude(P2, X2)

    # Place both cross-section Components and quickplot them
    c = gf.Component()
    c << wg1
    wg2ref = c << wg2
    wg2ref.movex(7.5)

    # Create the transitional CrossSection
    Xtrans = gf.path.transition(cross_section1=X1, cross_section2=X2, width_type="sine")
    # Create a Path for the transitional CrossSection to follow
    P3 = gf.path.straight(length=15, npoints=100)

    # Use the transitional CrossSection to create a Component
    straight_transition = gf.path.extrude(P3, Xtrans)

    P4 = gf.path.euler(radius=25, angle=45, p=0.5, use_eff=False)
    wg_trans = gf.path.extrude(P4, Xtrans)

    c = gf.Component()
    wg1_ref = c << wg1  # First cross-section Component
    wg2_ref = c << wg2
    wgt_ref = c << wg_trans

    wgt_ref.connect("o1", wg1_ref.ports["o2"])
    wg2_ref.connect("o1", wgt_ref.ports["o2"])

    wg3 = c << straight_transition
    wg3.movey(10)
    return c


if __name__ == "__main__":
    # c = test_path_transitions()
    # c= test_path_transition_function()
    c = test_path_transition_class()
    c.show(show_ports=True)

    # test_path_transition_class()

    # X1 = gf.cross_section.cross_section(width=1)
    # X2 = gf.cross_section.cross_section(width=3)
    # T = gf.path.transition(X1, X2)
    # P = gf.path.straight(length=10, npoints=101)
    # c3 = gf.path.extrude(P, T)

    # sections = {
    #     section["name"]: section for section in X1.sections if "name" in section
    # }

    # assert c3.ports["o1"].name == sections["_default"]["ports"][0]
    # assert c3.ports["o1"].layer == sections["_default"]["layer"]
    # assert c3.ports["o1"].orientation == 180
    # assert c3.ports["o1"].port_type == sections["_default"]["port_types"][0]
    # assert c3.ports["o1"].port_type == X1.port_type



from typing import List
from warnings import warn

import numpy as np

import gdsfactory as gf
from gdsfactory import Port
from gdsfactory.component import Component
from gdsfactory.components.wire import wire_corner
from gdsfactory.routing import get_route_from_waypoints
from gdsfactory.routing.manhattan import route_manhattan
from gdsfactory.types import CrossSectionSpec, LayerSpec, Route


class Node:
    def __init__(self, parent=None, position: tuple = ()):
        """Initializes a node. A node is a point on the grid."""
        self.parent = parent  # parent node of current node
        self.position = position  # position of current node

        self.g = 0  # distance between current node and start node
        self.h = 0  # distance between current node and end node
        self.f = self.g + self.h  # cost of the node (sum of g and h)


def get_route_astar(
    component: Component,
    port1: Port,
    port2: Port,
    resolution: float = 1,
    avoid_layers: List[LayerSpec] = None,
    distance: float = 1,
    cross_section: CrossSectionSpec = "strip",
    **kwargs,
) -> Route:
    """A* routing function. Finds a route between two ports avoiding obstacles.

    Args:
        component: Component the route, and ports belong to.
        port1: input.
        port2: output.
        resolution: discretization resolution in um.
            Lower resolution can help avoid accidental overlapping between route
            and components but adds more bends.
            The resolution decides how many "leaps/hops" the algorithm has to do.
        avoid_layers: list of layers to avoid.
        distance: distance from obstacles in um.
        cross_section: spec.
        kwargs: cross_section settings.
    """
    cross_section = gf.get_cross_section(cross_section, **kwargs)

    grid, x, y = _generate_grid(component, resolution, avoid_layers, distance)

    # Tell the algorithm which start and end directions to follow based on port orientation
    input_orientation = {
        0.0: (resolution, 0),
        90.0: (0, resolution),
        180.0: (-resolution, 0),
        270.0: (0, -resolution),
        None: (0, 0),
    }[port1.orientation]

    output_orientation = {
        0.0: (resolution, 0),
        90.0: (0, resolution),
        180.0: (-resolution, 0),
        270.0: (0, -resolution),
        None: (0, 0),
    }[port2.orientation]

    # Instantiate nodes
    start_node = Node(
        None,
        (
            round(port1.x + input_orientation[0]),
            round(port1.y + input_orientation[1]),
        ),
    )
    start_node.g = start_node.h = start_node.f = 0

    end_node = Node(
        None,
        (
            round(port2.x + output_orientation[0]),
            round(port2.y + output_orientation[1]),
        ),
    )
    end_node.g = end_node.h = end_node.f = 0

    # Add the start node
    open_list = [start_node]
    closed = []

    while open_list:
        # Current node
        current_index = 0
        for index in range(len(open_list)):
            if open_list[index].f < open_list[current_index].f:
                current_index = index

        # Pop current off open_list list, add to closed list
        current_node = open_list[current_index]
        closed.append(open_list.pop(current_index))

        # Reached end port
        if (
            current_node.position[0] == end_node.position[0]
            and current_node.position[1] == end_node.position[1]
        ):
            points = []
            current = current_node

            # trace back path from end node to start node
            while current is not None:
                points.append(current.position)
                current = current.parent
            # reverse to get true path
            points = points[::-1]

            # add the start and end ports
            points.insert(0, port1.center)
            points.append(port2.center)

            # return route from points
            if cross_section.radius:
                return get_route_from_waypoints(points, cross_section=cross_section)
            else:
                return get_route_from_waypoints(
                    points, cross_section=cross_section, bend=wire_corner
                )

        # Generate neighbours
        neighbours = _generate_neighbours(
            grid=grid,
            x=x,
            y=y,
            current_node=current_node,
            resolution=resolution,
        )

        # Loop through neighbours
        for neighbour in neighbours:

            for closed_neighbour in closed:
                if neighbour == closed_neighbour:
                    continue

            # Compute f, g, h
            neighbour.g = current_node.g + resolution
            # print(neighbour.g)
            neighbour.h = np.sqrt(
                (neighbour.position[0] - end_node.position[0]) ** 2
            ) + ((neighbour.position[1] - end_node.position[1]) ** 2)
            neighbour.f = neighbour.g + neighbour.h

            if current_node.parent is not None and (
                neighbour.position[0] - current_node.parent.position[0],
                neighbour.position[1] - current_node.parent.position[1],
            ) in [
                (resolution, -resolution),
                (-resolution, resolution),
                (resolution, resolution),
                (-resolution, -resolution),
            ]:
                neighbour.f *= 1.1  # penalize for turns

            # neighbour is already in the open_list
            for open_list_node in open_list:
                if neighbour == open_list_node and neighbour.g > open_list_node.g:
                    continue

            # Add the neighbour to open_list
            open_list.append(neighbour)

    warn("A* algorithm failed, resorting to Manhattan routing. Watch for overlaps.")
    return route_manhattan(port1, port2, cross_section=cross_section)


def _extract_all_bbox(c: Component, avoid_layers: List[LayerSpec] = None):
    """Extract all polygons whose layer is in `avoid_layers`."""
    return [c.get_polygons(layer) for layer in avoid_layers]


def _generate_grid(
    c: Component,
    resolution: float = 0.5,
    avoid_layers: List[LayerSpec] = None,
    distance: float = 1,
) -> np.ndarray:
    """Generate discretization grid that the algorithm will step through."""
    bbox = c.bbox
    x, y = np.meshgrid(
        np.linspace(
            bbox[0][0],
            bbox[1][0],
            int((bbox[1][0] - bbox[0][0]) / resolution),
            endpoint=True,
        ),
        np.linspace(
            bbox[0][1],
            bbox[1][1],
            int((bbox[1][1] - bbox[0][1]) / resolution),
            endpoint=True,
        ),
    )  # discretize component space
    x, y = x[0], y[:, 0]  # weed out copies
    grid = np.zeros(
        (len(x), len(y))
    )  # mapping from gdsfactory's x-, y- coordinate to grid vertex

    # assign 1 for obstacles
    if avoid_layers is None:
        for ref in c.references:
            bbox = ref.bbox
            xmin = np.abs(x - bbox[0][0] + distance).argmin()
            xmax = np.abs(x - bbox[1][0] - distance).argmin()
            ymin = np.abs(y - bbox[0][1] + distance).argmin()
            ymax = np.abs(y - bbox[1][1] - distance).argmin()

            grid[xmin:xmax, ymin:ymax] = 1
    else:
        all_refs = _extract_all_bbox(c, avoid_layers)
        for layer in all_refs:
            for bbox in layer:
                xmin = np.abs(x - bbox[0][0] + distance).argmin()
                xmax = np.abs(x - bbox[2][0] - distance).argmin()
                ymin = np.abs(y - bbox[0][1] + distance).argmin()
                ymax = np.abs(y - bbox[2][1] - distance).argmin()
                grid[xmin:xmax, ymin:ymax] = 1

    return np.ndarray.round(grid, 3), np.ndarray.round(x, 3), np.ndarray.round(y, 3)


def _generate_neighbours(
    current_node: Node,
    grid,
    x: np.ndarray,
    y: np.ndarray,
    resolution: float,
) -> List[Node]:
    """Generate neighbours of a node."""
    neighbours = []

    for new_position in [
        (0, -resolution),
        (0, resolution),
        (-resolution, 0),
        (resolution, 0),
    ]:  # Adjacent nodes along Manhattan path

        # Get node position
        node_position = (
            current_node.position[0] + new_position[0],
            current_node.position[1] + new_position[1],
        )

        # Make sure within range and not in obstacle
        if (
            node_position[0] > x.max()
            or node_position[0] < x.min()
            or node_position[1] > y.max()
            or node_position[1] < y.min()
        ):
            continue

        if (
            grid[
                next(
                    i
                    for i, _ in enumerate(x)
                    if np.isclose(_, node_position[0], atol=resolution)
                )
            ][
                next(
                    i
                    for i, _ in enumerate(y)
                    if np.isclose(_, node_position[1], atol=resolution)
                )
            ]
            == 1.0
        ):
            continue

        # Create new node
        new_node = Node(current_node, node_position)

        # Append
        neighbours.append(new_node)

    return neighbours


if __name__ == "__main__":

    cross_section = gf.get_cross_section("metal1", width=3)

    c = gf.Component("get_route_astar")
    w = gf.components.straight(cross_section=cross_section)

    left = c << w
    right = c << w
    right.move((100, 80))

    obstacle = gf.components.rectangle(size=(100, 3), layer="M1")
    obstacle1 = c << obstacle
    obstacle2 = c << obstacle
    obstacle1.ymin = 40
    obstacle2.xmin = 25

    port1 = left.ports["e2"]
    port2 = right.ports["e2"]

    routes = get_route_astar(
        component=c,
        port1=port1,
        port2=port2,
        cross_section=cross_section,
        resolution=5,
        distance=6.5,
    )

    c.add(routes.references)
    c.show()



from typing import Callable, List, Optional, Tuple, Union

import numpy as np
from numpy import float64, ndarray

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.components.via_corner import via_corner
from gdsfactory.components.wire import wire_corner
from gdsfactory.cross_section import strip
from gdsfactory.port import Port
from gdsfactory.routing.manhattan import (
    RouteError,
    get_route_error,
    remove_flat_angles,
    round_corners,
)
from gdsfactory.routing.path_length_matching import path_length_matched_points
from gdsfactory.routing.utils import get_list_ports_angle
from gdsfactory.types import (
    ComponentSpec,
    Coordinate,
    Coordinates,
    CrossSectionSpec,
    MultiCrossSectionAngleSpec,
    Number,
    Route,
)


def _is_vertical(segment: Coordinate, tol: float = 1e-5) -> bool:
    p0, p1 = segment
    return abs(p0[0] - p1[0]) < tol


def _is_horizontal(segment: Coordinate, tol: float = 1e-5) -> bool:
    p0, p1 = segment
    return abs(p0[1] - p1[1]) < tol


def _segment_sign(s: Coordinate) -> Number:
    p0, p1 = s
    if _is_vertical(s):
        return np.sign(p1[1] - p0[1])

    if _is_horizontal(s):
        return np.sign(p1[0] - p0[0])


def get_ports_x_or_y_distances(
    list_ports: List[Port], ref_point: Coordinate
) -> List[Number]:
    if not list_ports:
        return []

    angle = get_list_ports_angle(list_ports)
    x0 = ref_point[0]
    y0 = ref_point[1]
    return (
        [p.y - y0 for p in list_ports]
        if angle in [0, 180]
        else [p.x - x0 for p in list_ports]
    )


def _distance(port1, port2):
    x1, y1 = (port1.x, port1.y) if hasattr(port1, "x") else (port1[0], port1[1])
    x2, y2 = (port2.x, port2.y) if hasattr(port2, "x") else (port2[0], port2[1])
    dx = x1 - x2
    dy = y1 - y2

    return np.sqrt(dx**2 + dy**2)


def get_bundle_from_waypoints(
    ports1: List[Port],
    ports2: List[Port],
    waypoints: Coordinates,
    straight: ComponentSpec = straight_function,
    taper: ComponentSpec = taper_function,
    bend: ComponentSpec = bend_euler,
    sort_ports: bool = True,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = strip,
    separation: Optional[float] = None,
    on_route_error: Callable = get_route_error,
    path_length_match_loops: Optional[int] = None,
    path_length_match_extra_length: float = 0.0,
    path_length_match_modify_segment_i: int = -2,
    **kwargs,
) -> List[Route]:
    """Returns list of routes that connect bundle of ports with bundle of routes.

    Routes follow a list of waypoints.
    Take a look at get_bundle_from_steps for easier definition.

    Args:
        ports1: list of ports.
        ports2: list of ports.
        waypoints: list of points defining a route.
        straight: function that returns straights.
        taper: function that returns tapers.
        bend: function that returns bends.
        sort_ports: sorts ports.
        cross_section: cross_section.
        separation: center to center, defaults to ports1 separation.
        on_route_error: function to call for routing errors.
        path_length_match_loops: Integer number of loops to add to bundle
            for path length matching (won't try to match if None).
        path_length_match_extra_length: Extra length to add
            to path length matching loops (requires path_length_match_loops != None).
        path_length_match_modify_segment_i: Index of straight segment to add path
            length matching loops to (requires path_length_match_loops != None).
        kwargs: cross_section settings.

    """
    if len(ports2) != len(ports1):
        raise ValueError(
            f"Number of start ports should match number of end ports.\
        Got {len(ports1)} {len(ports2)}"
        )

    waypoints = [ports1[0].center] + list(waypoints) + [ports2[0].center]
    for p in ports1:
        # if ports1 orientation is None, guess it from the first two waypoints
        if p.orientation is None:
            if np.subtract(waypoints[1], waypoints[0])[0] > 0:
                p.orientation = 0
            elif np.subtract(waypoints[1], waypoints[0])[0] < 0:
                p.orientation = 180
            elif np.subtract(waypoints[1], waypoints[0])[1] > 0:
                p.orientation = 90
            elif np.subtract(waypoints[1], waypoints[0])[1] < 0:
                p.orientation = 270
        p.orientation = int(p.orientation) % 360 if p.orientation else p.orientation

    for p in ports2:
        # if ports2 orientation is None, guess it from the last two waypoints
        if p.orientation is None:
            if np.subtract(waypoints[-1], waypoints[-2])[0] > 0:
                p.orientation = 180
            elif np.subtract(waypoints[-1], waypoints[-2])[0] < 0:
                p.orientation = 0
            elif np.subtract(waypoints[-1], waypoints[-2])[1] > 0:
                p.orientation = 90
            elif np.subtract(waypoints[-1], waypoints[-2])[1] < 0:
                p.orientation = 270
        p.orientation = int(p.orientation) % 360 if p.orientation else p.orientation

    if sort_ports:
        # Sort the ports such that the bundle connect the correct corresponding ports.
        angles_to_sorttypes = {
            (0, 180): ("Y", "Y"),
            (0, 90): ("Y", "X"),
            (0, 0): ("Y", "-Y"),
            (0, 270): ("Y", "-X"),
            (90, 0): ("X", "Y"),
            (90, 90): ("X", "-X"),
            (90, 180): ("X", "-Y"),
            (90, 270): ("X", "X"),
            (180, 90): ("Y", "-X"),
            (180, 0): ("Y", "Y"),
            (180, 270): ("Y", "X"),
            (180, 180): ("Y", "-Y"),
            (270, 90): ("X", "X"),
            (270, 270): ("X", "-X"),
            (270, 0): ("X", "-Y"),
            (270, 180): ("X", "Y"),
        }

        dict_sorts = {
            "X": lambda p: p.x,
            "Y": lambda p: p.y,
            "-X": lambda p: -p.x,
            "-Y": lambda p: -p.y,
        }
        start_angle = ports1[0].orientation
        end_angle = ports2[0].orientation

        key = (start_angle, end_angle)
        sp_st, ep_st = angles_to_sorttypes[key]
        start_port_sort = dict_sorts[sp_st]
        end_port_sort = dict_sorts[ep_st]
        ports1.sort(key=start_port_sort)
        ports2.sort(key=end_port_sort)

    try:
        routes = _generate_manhattan_bundle_waypoints(
            ports1=ports1,
            ports2=ports2,
            waypoints=list(waypoints),
            separation=separation,
            **kwargs,
        )
    except RouteError:
        return [on_route_error(waypoints)]

    # bends90 = [
    #     gf.get_component(bend, cross_section=cross_section, **kwargs) for p in ports1
    # ]

    if taper and not isinstance(cross_section, list):
        x = gf.get_cross_section(cross_section, **kwargs)
        if x.auto_widen and callable(taper):
            taper = gf.get_component(
                taper,
                length=x.taper_length,
                width1=ports1[0].width,
                width2=x.width_wide,
                layer=ports1[0].layer,
            )
    else:
        taper = None

    if path_length_match_loops:
        routes = [np.array(route) for route in routes]
        routes = path_length_matched_points(
            routes,
            extra_length=path_length_match_extra_length,
            bend=bend,
            nb_loops=path_length_match_loops,
            modify_segment_i=path_length_match_modify_segment_i,
            cross_section=cross_section,
            **kwargs,
        )
    return [
        round_corners(
            points=pts,
            bend=bend,
            straight=straight,
            taper=taper,
            cross_section=cross_section,
            **kwargs,
        )
        for pts in routes
    ]


get_bundle_from_waypoints_electrical = gf.partial(
    get_bundle_from_waypoints, bend=wire_corner, cross_section="metal_routing"
)

get_bundle_from_waypoints_electrical_multilayer = gf.partial(
    get_bundle_from_waypoints_electrical,
    bend=via_corner,
    cross_section=[
        (gf.cross_section.metal2, (90, 270)),
        ("metal_routing", (0, 180)),
    ],
)


def snap_route_to_end_point_x(route, x):
    y1, y2 = (p[1] for p in route[-2:])
    return route[:-2] + [(x, y1), (x, y2)]


def snap_route_to_end_point_y(
    route: List[Union[ndarray, Tuple[float64, float64]]], y: float64
) -> List[Union[ndarray, Tuple[float64, float64]]]:
    x1, x2 = (p[0] for p in route[-2:])
    return route[:-2] + [(x1, y), (x2, y)]


def _generate_manhattan_bundle_waypoints(
    ports1: List[Port],
    ports2: List[Port],
    waypoints: Coordinates,
    separation: Optional[float] = None,
    **kwargs,
) -> Coordinates:
    """Returns waypoints for bundle.

    Args:
        ports1: list of ports must face the same direction.
        ports2: list of ports must face the same direction.
        waypoints: from one point within the ports1 bank
            to another point within the ports2 bank.
        separation: center to center, defaults to ports1 separation.
    """
    waypoints = remove_flat_angles(waypoints)
    way_segments = list(zip(waypoints, waypoints[1:]))
    offsets_start = get_ports_x_or_y_distances(ports1, waypoints[0])

    start_angle = ports1[0].orientation
    if start_angle in [90]:
        offsets_start = [-_d for _d in offsets_start]
    end_angle = ports2[0].orientation

    # if separation is defined, the offsets should increment from the reference port
    # in the same direction as the original offsets
    if separation and len(ports1) > 1:
        # the default case when we start with the reference port
        offsets_start_sign = (
            np.sign(offsets_start[1]) if np.sign(offsets_start[1]) != 0 else 1
        )
        offsets_mid = [
            offsets_start_sign * separation * i for i, o in enumerate(offsets_start)
        ]
        if offsets_start[0] == 0:
            # the default case, no action necessary
            pass
        elif offsets_start[-1] == 0:
            # if the reference port is last, reverse the whole list
            offsets_mid.reverse()
        else:
            raise ValueError("Expected offset = 0 at either start or end of route.")

    # if there is only one route, we should skip this step as it is irrelevant
    else:
        # separation defaults to ports1 separation
        offsets_mid = offsets_start

    def _displace_segment_copy(s, a, sh=1, sv=1):
        sign_seg = _segment_sign(s)
        if _is_horizontal(s):
            dp = (0, sh * sign_seg * a)
        elif _is_vertical(s):
            dp = (sv * sign_seg * a, 0)
        else:
            raise RouteError(f"Segment should be manhattan, got {s}")

        return [np.array(p) + dp for p in s]

    def _displace_segment_copy_group1(s, a):
        return _displace_segment_copy(s, a, sh=1, sv=-1)

    def _intersection(s1, s2):
        if _is_horizontal(s1) and _is_vertical(s2):
            sh, sv = s1, s2
        elif _is_horizontal(s2) and _is_vertical(s1):
            sh, sv = s2, s1
        else:

            if _is_horizontal(s1):
                s1_dir = "h"
            elif _is_vertical(s1):
                s1_dir = "v"
            else:
                s1_dir = "u"

            if _is_horizontal(s2):
                s2_dir = "h"
            elif _is_vertical(s2):
                s2_dir = "v"
            else:
                s2_dir = "u"

            raise ValueError(
                f"s1 / s2 should be h/v or v/h. Got {s1_dir} {s2_dir} {s1} {s2}"
            )
        return sv[0][0], sh[0][1]

    routes = []
    _make_segment = _displace_segment_copy_group1
    for i, start_port in enumerate(ports1):
        route = []
        prev_seg_sep = None

        for j, seg in enumerate(way_segments):
            if j == 0:
                start_point = start_port.center
                seg_sep = offsets_start[i]
            else:
                seg_sep = offsets_mid[i]
                d_seg = _make_segment(seg, seg_sep)
                prev_seg = way_segments[j - 1]
                tmp_seg = _make_segment(prev_seg, prev_seg_sep)
                start_point = _intersection(d_seg, tmp_seg)

            route += [start_point]

            # If last point before the ports, adjust the separation to the end ports
            if j == len(way_segments) - 1:
                end_point = ports2[i].center
                route += [end_point]

                if end_angle in [0, 180]:
                    route = snap_route_to_end_point_y(route, end_point[1])
                else:
                    route = snap_route_to_end_point_x(route, end_point[0])
            prev_seg_sep = seg_sep
        routes += [route]

    return routes



from typing import Callable


class Event:
    def __init__(self):
        """Initialize the object."""
        self._handlers = []

    def __iadd__(self, handler: Callable):
        """Add item."""
        self.add_handler(handler)
        return self

    def __isub__(self, handler: Callable):
        """Subtract item."""
        self._handlers.remove(handler)
        return self

    def add_handler(self, handler: Callable):
        """Adds a handler that will be executed when this event is fired.

        Args:
            handler: a function which matches the signature fired by this event

        """
        self._handlers.append(handler)

    def clear_handlers(self):
        """Clear all handlers."""
        self._handlers.clear()

    def fire(self, *args, **kwargs):
        """Fires an event, calling all handlers with the passed arguments."""
        for eventhandler in self._handlers:
            eventhandler(*args, **kwargs)



"""snaps values and coordinates to the GDS grid in nm."""
from typing import Optional, Tuple, Union

import numpy as np


def is_on_grid(x: float, nm: int = 1) -> bool:
    return np.array_equal(snap_to_grid(x, nm=nm), np.round(x, 6))


def assert_on_1nm_grid(x: float) -> None:
    x_grid = snap_to_grid(x)
    if not np.isclose(x_grid, x):
        raise ValueError(f"{x} needs to be on 1nm grid, try {x_grid}")


def assert_on_2nm_grid(x: float) -> None:
    x_grid = snap_to_2nm_grid(x)
    if not np.isclose(x_grid, x):
        raise ValueError(f"{x} needs to be on 2nm grid, try {x_grid}")


def snap_to_grid(
    x: Union[float, Tuple, np.ndarray], nm: Optional[int] = None
) -> Union[float, Tuple, np.ndarray]:

    if nm is None:
        from gdsfactory.pdk import get_grid_size

        nm = int(get_grid_size() * 1000)
    elif nm == 0:
        return x
    elif nm < 0:
        raise ValueError("nm must be an integer tolerance value greater than zero")

    if nm == 1:
        y = np.round(np.asarray(x, dtype=float), 3)
    else:
        y = nm * np.round(np.asarray(x, dtype=float) * 1e3 / nm) / 1e3
    if isinstance(x, tuple):
        return tuple(y)
    elif isinstance(x, (int, float, str, np.float_)):
        return float(y)
    return y


def snap_to_2nm_grid(x: float) -> float:
    return snap_to_grid(x, nm=2)


def snap_to_5nm_grid(x: float) -> float:
    return snap_to_grid(x, nm=5)


if __name__ == "__main__":
    print(snap_to_grid(1.1e-3))
    # print(snap_to_2nm_grid(1.1e-3))
    # print(snap_to_2nm_grid(3.1e-3))

    # print(on_1nm_grid(1.1e-3))
    # print(on_1nm_grid(1e-3))

    # print(on_2nm_grid(1.1e-3))
    # print(on_2nm_grid(1e-3))
    # print(on_2nm_grid(2e-3))



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component


def demo_get_bundle_sbend_routing(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:
    """FIXME."""

    lengths = {}

    c = gf.Component("test_get_bundle_sort_ports")
    pitch = 2.0
    ys_left = [0, 10, 20]
    N = len(ys_left)
    ys_right = [(i - N / 2) * pitch for i in range(N)]

    right_ports = [
        gf.Port(f"R_{i}", center=(0, ys_right[i]), width=0.5, orientation=180)
        for i in range(N)
    ]
    left_ports = [
        gf.Port(f"L_{i}", center=(-50, ys_left[i]), width=0.5, orientation=0)
        for i in range(N)
    ]
    left_ports.reverse()
    routes = gf.routing.get_bundle(right_ports, left_ports, bend_radius=5)

    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = demo_get_bundle_sbend_routing(None, check=False)
    c.show(show_ports=True)



import gdspy

import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def xor_diff(A, B, precision: float = 1e-4) -> Component:
    """Given two Components A and B, performs the layer-by-layer XOR difference \
    between A and B and returns polygons representing the differences between A \
    and B.

    gdsfactory wrapper for phidl.geometry.xor_diff

    Args:
        A: Component(/Reference) or list of Component(/References).
        B: Component(/Reference) or list of Component(/References).
        precision: Desired precision for rounding vertex coordinates.

    Returns
        Component: containing a polygon(s) defined by the XOR difference result
        between A and B.

    """
    D = Component()
    A_polys = A.get_polygons(by_spec=True)
    B_polys = B.get_polygons(by_spec=True)
    A_layers = A_polys.keys()
    B_layers = B_polys.keys()
    all_layers = set()
    all_layers.update(A_layers)
    all_layers.update(B_layers)
    for layer in all_layers:
        if (layer in A_layers) and (layer in B_layers):
            p = gdspy.boolean(
                operand1=A_polys[layer],
                operand2=B_polys[layer],
                operation="xor",
                precision=precision,
                max_points=4000,
                layer=layer[0],
                datatype=layer[1],
            )
        elif layer in A_layers:
            p = A_polys[layer]
        elif layer in B_layers:
            p = B_polys[layer]
        if p is not None:
            D.add_polygon(p, layer=layer)
    return D


if __name__ == "__main__":
    e1 = gf.components.ellipse(radii=(6, 6))
    e2 = gf.components.ellipse(radii=(10, 4))
    c = xor_diff(A=e1, B=e2)
    c.show(show_ports=True)



"""Tidy3D."""

import concurrent.futures
import hashlib

import tidy3d as td
from tidy3d import web

import gdsfactory as gf
from gdsfactory.config import PATH
from gdsfactory.types import List, PathType

_executor = concurrent.futures.ThreadPoolExecutor()


def get_sim_hash(sim: td.Simulation) -> str:
    """Returns simulation hash as the unique ID."""
    return hashlib.md5(str(sim).encode()).hexdigest()


def _get_results(
    sims: List[td.Simulation],
    dirpath: PathType = PATH.results_tidy3d,
    overwrite: bool = False,
) -> web.BatchData:
    """Return SimulationData results from simulation.

    Only submits simulation if results not found locally or remotely.
    First tries to load simulation results from disk.
    Then it tries to load them from the server storage.
    Finally, submits simulation to run remotely

    Args:
        sim: tidy3d Simulation.
        dirpath: to store results locally.
        overwrite: overwrites the data even when path exists.

    """
    # task_name = sim_hash = get_sim_hash(sim)
    # sim_path = dirpath / f"{sim_hash}.hdf5"
    # logger.info(f"running simulation {sim_hash!r}")

    # hash_to_id = {d["task_name"][:32]: d["task_id"] for d in web.get_tasks()}
    # filepath = str(dirpath / f"{sim_hash}.hdf5")
    # job = web.Job(simulation=sim, task_name=task_name)

    # # Results in local storage
    # if sim_path.exists():
    #     task_id = hash_to_id[sim_hash]
    #     logger.info(f"{sim_path!r} for task_id {task_id!r} found in local storage")
    #     return td.SimulationData.from_file(filepath)

    # # Results in server storage
    # if sim_hash in hash_to_id:
    #     task_id = hash_to_id[sim_hash]
    #     web.monitor(task_id)

    #     try:
    #         return web.load(task_id=task_id, path=filepath, replace_existing=overwrite)
    #     except WebError:
    #         logger.info(f"task_id {task_id!r} exists but no results found.")
    #     except Exception:
    #         logger.info(f"task_id {task_id!r} exists but unexpected error encountered.")

    # # Run simulation if results not found in local or server storage
    # logger.info(f"sending task_name {task_name!r} to tidy3d server.")
    # return job.run(path=filepath)

    task_names = [get_sim_hash(sim) for sim in sims]
    batch = web.Batch(simulations=dict(zip(task_names, sims)))
    return batch.run(path_dir=dirpath)


def get_results(
    sims: List[td.Simulation],
    dirpath=PATH.results_tidy3d,
    overwrite: bool = True,
) -> web.BatchData:
    """Return a List of SimulationData from a list of Simulation.

    Works with Pool of threads.
    Each thread can run in parallel and only becomes blocking when you ask
    for .result()

    Args:
        sims: List[Simulation]
        dirpath: to store results locally
        overwrite: overwrites the data even if path exists. Keep True.


    .. code::
        import gdsfactory.simulation.tidy3d as gt

        component = gf.components.straight(length=3)
        sim = gt.get_simulation(component=component)
        sim_data = gt.get_results([sim])

    """
    return _get_results(sims, dirpath, overwrite)


if __name__ == "__main__":
    import gdsfactory.simulation.gtidy3d as gt

    component = gf.components.straight(length=3)
    sim = gt.get_simulation(component=component)
    r = sim_data = get_results(sim=sim).result()



from gdsfactory.read.from_dphox import from_dphox
from gdsfactory.read.from_gdspaths import from_gdsdir, from_gdspaths
from gdsfactory.read.from_np import from_np
from gdsfactory.read.from_phidl import from_gdspy, from_phidl
from gdsfactory.read.from_picwriter import from_picwriter
from gdsfactory.read.from_yaml import from_yaml
from gdsfactory.read.import_gds import import_gds
from gdsfactory.read.import_oas import import_oas

__all__ = [
    "from_dphox",
    "from_gdsdir",
    "from_gdspaths",
    "from_np",
    "from_phidl",
    "from_picwriter",
    "from_yaml",
    "from_gdspy",
    "import_gds",
    "import_oas",
]



import gdsfactory as gf
from gdsfactory.pdk import get_grid_size
from gdsfactory.snap import is_on_grid
from gdsfactory.types import Component, ComponentReference, Optional


def is_valid_transformation(
    ref: ComponentReference, grid_size: Optional[float] = None
) -> bool:
    """Returns True if the component has valid transformations.

    Args:
        component: the component reference to check.
        grid_size: the GDS grid size, in um, defaults to active PDK.get_grid_size()
            any translations with higher resolution than this are considered invalid.
    """
    grid_size = grid_size or get_grid_size()
    nm = int(grid_size * 1e3)
    origin_is_on_grid = all(is_on_grid(x, nm) for x in ref.origin)
    rotation_is_regular = ref.rotation is None or ref.rotation % 90 == 0
    return origin_is_on_grid and rotation_is_regular


def has_valid_transformations(component: Component) -> bool:
    """Returns True if the component has valid transformations."""
    refs = component.references
    return all(is_valid_transformation(ref) for ref in refs)


def flatten_invalid_refs(component: Component, grid_size: Optional[float] = None):
    """Flattens component references component with invalid GDS.

    transformations.

    (i.e. non-90 deg rotations or sub-grid translations).

    This is an in-place operation, so you should use it as a decorator.
    flattens only individual references with invalid transformations.

    Args:
        component: the component to fix (in place).
        grid_size: the GDS grid size, in um, defaults to active PDK.get_grid_size()
            any translations with higher resolution than this are considered invalid.
    """
    grid_size = grid_size or get_grid_size()
    nm = int(grid_size * 1e3)

    refs = component.references.copy()
    for ref in refs:
        origin_is_on_grid = all(is_on_grid(x, nm) for x in ref.origin)
        rotation_is_regular = ref.rotation is None or ref.rotation % 90 == 0
        if not origin_is_on_grid or not rotation_is_regular:
            component.flatten_reference(ref)
    return component


@gf.cell
def _demo_non_manhattan() -> Component:
    """Returns component with Manhattan snapping issues."""
    c = Component()
    b = c << gf.components.bend_circular(angle=30)
    s = c << gf.components.straight(length=5)
    s.connect("o1", b.ports["o2"])
    return c


def test_flatten_invalid_refs():
    c1 = _demo_non_manhattan()
    assert not has_valid_transformations(c1)

    c2 = _demo_non_manhattan(decorator=flatten_invalid_refs)
    assert has_valid_transformations(c2)


if __name__ == "__main__":
    test_flatten_invalid_refs()
    # c = _demo_non_manhattan(decorator=flatten_invalid_refs)
    # c.show()



"""Compute and write Sparameters using Meep in MPI."""

import multiprocessing
import pathlib
import pickle
import shlex
import subprocess
import sys
import time
from pathlib import Path
from typing import Optional

import pydantic

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.config import logger, sparameters_path
from gdsfactory.pdk import get_layer_stack
from gdsfactory.simulation import port_symmetries
from gdsfactory.simulation.get_sparameters_path import (
    get_sparameters_path_meep as get_sparameters_path,
)
from gdsfactory.simulation.gmeep.write_sparameters_meep import (
    remove_simulation_kwargs,
    settings_write_sparameters_meep,
)
from gdsfactory.tech import LayerStack
from gdsfactory.types import ComponentSpec, PathType

ncores = multiprocessing.cpu_count()

temp_dir_default = Path(sparameters_path) / "temp"


def _python() -> str:
    """Select correct python executable from current activated environment."""
    return sys.executable


@pydantic.validate_arguments
def write_sparameters_meep_mpi(
    component: ComponentSpec,
    layer_stack: Optional[LayerStack] = None,
    cores: int = ncores,
    filepath: Optional[PathType] = None,
    dirpath: Optional[PathType] = None,
    temp_dir: Path = temp_dir_default,
    temp_file_str: str = "write_sparameters_meep_mpi",
    live_output: bool = False,
    overwrite: bool = False,
    wait_to_finish: bool = True,
    **kwargs,
) -> Path:
    """Write Sparameters using multiple cores and MPI and returns Sparameters filepath.

    Simulates each time using a different input port (by default, all of them)
    unless you specify port_symmetries:

    checks stderror and kills MPI job if there is any stderror message

    port_symmetries = {"o1":
            {
                "s11": ["s22","s33","s44"],
                "s21": ["s21","s34","s43"],
                "s31": ["s13","s24","s42"],
                "s41": ["s14","s23","s32"],
            }
        }

    Args:
        component: gdsfactory Component.
        cores: number of processors.
        filepath: to store pandas Dataframe with Sparameters in CSV format.
            Defaults to dirpath/component_.csv.
        dirpath: directory to store sparameters in CSV.
            Defaults to active Pdk.sparameters_path.
        layer_stack: contains layer to thickness, zmin and material.
            Defaults to active pdk.layer_stack.
        temp_dir: temporary directory to hold simulation files.
        temp_file_str: names of temporary files in temp_dir.
        live_output: stream output of mpirun command to file and print to console
            (meep verbosity still needs to be set separately).
        overwrite: overwrites stored simulation results.
        wait_to_finish: if True makes the function call blocking.

    Keyword Args:
        resolution: in pixels/um (30: for coarse, 100: for fine).
        port_symmetries: Dict to specify port symmetries, to save number of simulations.
        dirpath: directory to store Sparameters.
        port_margin: margin on each side of the port.
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port.
        port_source_offset: offset between source GDS port and source MEEP port.
        filepath: to store pandas Dataframe with Sparameters in CSV format.
        animate: saves a MP4 images of the simulation for inspection, and also
            outputs during computation. The name of the file is the source index.
        lazy_parallelism: toggles the flag "meep.divide_parallel_processes" to
            perform the simulations with different sources in parallel.
        dispersive: use dispersive models for materials (requires higher resolution).
        xmargin: left and right distance from component to PML.
        xmargin_left: west distance from component to PML.
        xmargin_right: east distance from component to PML.
        ymargin: top and bottom distance from component to PML.
        ymargin_top: north distance from component to PML.
        ymargin_bot: south distance from component to PML.
        extend_ports_length: to extend ports beyond the PML.
        zmargin_top: thickness for cladding above core.
        zmargin_bot: thickness for cladding below core.
        tpml: PML thickness (um).
        clad_material: material for cladding.
        is_3d: if True runs in 3D.
        wavelength_start: wavelength min (um).
        wavelength_stop: wavelength max (um).
        wavelength_points: wavelength steps.
        dfcen: delta frequency.
        port_source_name: input port name.
        port_margin: margin on each side of the port.
        distance_source_to_monitors: in (um) source goes before.
        port_source_offset: offset between source GDS port and source MEEP port.
        port_monitor_offset: offset between monitor GDS port and monitor MEEP port.

    Returns:
        filepath for sparameters CSV (wavelengths, s11a, o1@0,o2@0, ...)
            where `a` is the angle in radians and `m` the module.

    TODO:
        write stdout to file, maybe simulation logs too.

    """
    for setting in kwargs:
        if setting not in settings_write_sparameters_meep:
            raise ValueError(f"{setting!r} not in {settings_write_sparameters_meep}")

    component = gf.get_component(component)
    assert isinstance(component, Component)

    layer_stack = layer_stack or get_layer_stack()

    settings = remove_simulation_kwargs(kwargs)
    filepath = filepath or get_sparameters_path(
        component=component,
        dirpath=dirpath,
        layer_stack=layer_stack,
        **settings,
    )
    filepath = pathlib.Path(filepath)
    if filepath.exists() and not overwrite:
        logger.info(f"Simulation {filepath!r} already exists")
        return filepath

    if filepath.exists() and overwrite:
        filepath.unlink()

    # Save all the simulation arguments for later retrieval
    temp_dir.mkdir(exist_ok=True, parents=True)
    tempfile = temp_dir / temp_file_str
    parameters_file = tempfile.with_suffix(".pkl")
    kwargs.update(filepath=str(filepath))

    parameters_dict = {
        "component": component,
        "layer_stack": layer_stack,
        "overwrite": overwrite,
    }

    # Loop over kwargs
    for key in kwargs:
        parameters_dict[key] = kwargs[key]

    with open(parameters_file, "wb") as outp:
        pickle.dump(parameters_dict, outp, pickle.HIGHEST_PROTOCOL)

    # Write execution file
    script_lines = [
        "import pickle\n",
        "from gdsfactory.simulation.gmeep import write_sparameters_meep\n\n",
        'if __name__ == "__main__":\n\n',
        f"\twith open(\"{parameters_file}\", 'rb') as inp:\n",
        "\t\tparameters_dict = pickle.load(inp)\n\n" "\twrite_sparameters_meep(\n",
    ]
    script_lines.extend(
        f'\t\t{key} = parameters_dict["{key}"],\n' for key in parameters_dict
    )

    script_lines.append("\t)")
    script_file = tempfile.with_suffix(".py")
    with open(script_file, "w") as script_file_obj:
        script_file_obj.writelines(script_lines)
    command = f"mpirun -np {cores} {_python()} {script_file}"
    logger.info(command)
    logger.info(str(filepath))

    if live_output:
        import asyncio

        from gdsfactory.async_utils import execute_and_stream_output

        asyncio.run(
            execute_and_stream_output(
                command, log_file_dir=temp_dir, log_file_str=temp_file_str
            )
        )
    else:
        with subprocess.Popen(
            shlex.split(command),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            print(proc.stdout.read().decode())
            print(proc.stderr.read().decode())
            sys.stdout.flush()
            sys.stderr.flush()
        if wait_to_finish and not proc.stderr:
            while not filepath.exists():
                print(proc.stdout.read().decode())
                print(proc.stderr.read().decode())
                sys.stdout.flush()
                sys.stderr.flush()
                time.sleep(1)
    return filepath


write_sparameters_meep_mpi_1x1 = gf.partial(
    write_sparameters_meep_mpi, port_symmetries=port_symmetries.port_symmetries_1x1
)

write_sparameters_meep_mpi_1x1_bend90 = gf.partial(
    write_sparameters_meep_mpi,
    ymargin_bot=3,
    ymargin=0,
    xmargin_right=3,
    port_symmetries=port_symmetries.port_symmetries_1x1,
)


if __name__ == "__main__":
    c1 = gf.components.straight(length=2.1)
    filepath = write_sparameters_meep_mpi(
        component=c1,
        # ymargin=3,
        cores=2,
        run=True,
        overwrite=True,
        live_output=True,
        # lazy_parallelism=True,
        lazy_parallelism=False,
        # filepath="instance_dict.csv",
    )



import typing
import warnings
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import numpy as np
from gdspy import CellReference
from numpy import cos, float64, int64, mod, ndarray, pi, sin

from gdsfactory.component_layout import _GeometryHelper
from gdsfactory.port import (
    Port,
    map_ports_layer_to_orientation,
    map_ports_to_orientation_ccw,
    map_ports_to_orientation_cw,
    select_ports,
)
from gdsfactory.snap import snap_to_grid

if typing.TYPE_CHECKING:
    from gdsfactory.component import Component

Number = Union[float64, int64, float, int]
Coordinate = Union[Tuple[Number, Number], ndarray, List[Number]]
Coordinates = Union[List[Coordinate], ndarray, List[Number], Tuple[Number, ...]]


class SizeInfo:
    def __init__(self, bbox: ndarray) -> None:
        """Initialize this object."""
        self.west = bbox[0, 0]
        self.east = bbox[1, 0]
        self.south = bbox[0, 1]
        self.north = bbox[1, 1]

        self.width = self.east - self.west
        self.height = self.north - self.south

        xc = 0.5 * (self.east + self.west)
        yc = 0.5 * (self.north + self.south)

        self.sw = np.array([self.west, self.south])
        self.se = np.array([self.east, self.south])
        self.nw = np.array([self.west, self.north])
        self.ne = np.array([self.east, self.north])

        self.cw = np.array([self.west, yc])
        self.ce = np.array([self.east, yc])
        self.nc = np.array([xc, self.north])
        self.sc = np.array([xc, self.south])
        self.cc = self.center = np.array([xc, yc])

    def get_rect(
        self, padding=0, padding_w=None, padding_e=None, padding_n=None, padding_s=None
    ) -> Tuple[Coordinate, Coordinate, Coordinate, Coordinate]:
        w, e, s, n = self.west, self.east, self.south, self.north

        padding_n = padding if padding_n is None else padding_n
        padding_e = padding if padding_e is None else padding_e
        padding_w = padding if padding_w is None else padding_w
        padding_s = padding if padding_s is None else padding_s

        w = w - padding_w
        e = e + padding_e
        s = s - padding_s
        n = n + padding_n

        return ((w, s), (e, s), (e, n), (w, n))

    @property
    def rect(self) -> Tuple[Coordinate, Coordinate]:
        return self.get_rect()

    def __str__(self) -> str:
        """Return a string representation of the object."""
        return f"w: {self.west}\ne: {self.east}\ns: {self.south}\nn: {self.north}\n"


def _rotate_points(
    points: Coordinates,
    angle: float = 45.0,
    center: Coordinate = (
        0.0,
        0.0,
    ),
) -> ndarray:
    """Rotates points around a center point.

    accepts single points [1,2] or array-like[N][2], and will return in kind

    Args:
        points: rotate points around center point.
        angle: in degrees.
        center: x, y.
    """
    # First check for common, easy values of angle
    p_arr = np.asarray(points)
    if angle == 0:
        return p_arr

    c0 = np.asarray(center)
    displacement = p_arr - c0
    if angle == 180:
        return c0 - displacement

    if p_arr.ndim == 2:
        perpendicular = displacement[:, ::-1]
    elif p_arr.ndim == 1:
        perpendicular = displacement[::-1]

    # Fall back to trigonometry
    angle = angle * pi / 180
    ca = cos(angle)
    sa = sin(angle)
    sa = np.array((-sa, sa))
    return displacement * ca + perpendicular * sa + c0


class ComponentReference(CellReference, _GeometryHelper):
    """Pointer to a Component with x, y, rotation, mirror."""

    def __init__(
        self,
        component: "Component",
        origin: Coordinate = (0, 0),
        rotation: float = 0,
        magnification: None = None,
        x_reflection: bool = False,
        visual_label: str = "",
    ) -> None:
        """Initialize the ComponentReference object."""
        CellReference.__init__(
            self,
            ref_cell=component,
            origin=origin,
            rotation=rotation,
            magnification=magnification,
            x_reflection=x_reflection,
            ignore_missing=False,
        )
        self._owner = None
        self._name = None

        # The ports of a ComponentReference have their own unique id (uid),
        # since two ComponentReferences of the same parent Component can be
        # in different locations and thus do not represent the same port
        self._local_ports = {
            name: port._copy() for name, port in component.ports.items()
        }
        self.visual_label = visual_label
        # self.uid = str(uuid.uuid4())[:8]

    @property
    def parent(self):
        return self.ref_cell

    @parent.setter
    def parent(self, value):
        self.ref_cell = value

    @property
    def owner(self):
        return self._owner

    @owner.setter
    def owner(self, value):
        if self.owner is None or value is None:
            self._owner = value
        elif value != self._owner:
            raise ValueError(
                f"Cannot reset owner of a reference once it has already been set!"
                f" Reference: {self}. Current owner: {self._owner}. "
                f"Attempting to re-assign to {value!r}"
            )

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if value != self._name:
            if self.owner and value in self.owner.named_references:
                raise ValueError(
                    f"This reference's owner already has a reference with name {value!r}. Please choose another name."
                )
            self._name = value
            self.owner._reference_names_used.add(value)

    @property
    def alias(self):
        warnings.warn(
            "alias attribute is deprecated and may be removed in a future version of gdsfactory",
            DeprecationWarning,
        )
        return self.name

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return (
            'ComponentReference (parent Component "%s", ports %s, origin %s, rotation %s,'
            " x_reflection %s)"
            % (
                self.parent.name,
                list(self.ports.keys()),
                self.origin,
                self.rotation,
                self.x_reflection,
            )
        )

    def to_dict(self):
        d = self.parent.to_dict()
        d.update(
            origin=self.origin,
            rotation=self.rotation,
            magnification=self.magnification,
            x_reflection=self.x_reflection,
        )
        return d

    @property
    def bbox(self):
        """Return the bounding box of the ComponentReference.

        it snaps to 3 decimals in um (0.001um = 1nm precision)
        """
        bbox = self.get_bounding_box()
        if bbox is None:
            bbox = ((0, 0), (0, 0))
        return np.round(bbox, 3)

    @classmethod
    def __get_validators__(cls):
        """Get validators."""
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """Check with pydantic ComponentReference valid type."""
        assert isinstance(
            v, ComponentReference
        ), f"TypeError, Got {type(v)}, expecting ComponentReference"
        return v

    def __getitem__(self, val):
        """This allows you to access an alias from the reference's parent, and.

        receive a copy of the reference which is correctly rotated and
        translated.
        """
        try:
            alias_device = self.parent[val]
        except Exception as exc:
            raise ValueError(
                '[PHIDL] Tried to access alias "%s" from parent '
                'Component "%s", which does not exist' % (val, self.parent.name)
            ) from exc
        new_reference = ComponentReference(
            alias_device.parent,
            origin=alias_device.origin,
            rotation=alias_device.rotation,
            magnification=alias_device.magnification,
            x_reflection=alias_device.x_reflection,
        )

        if self.x_reflection:
            new_reference.reflect((1, 0))
        if self.rotation is not None:
            new_reference.rotate(self.rotation)
        if self.origin is not None:
            new_reference.move(self.origin)

        return new_reference

    @property
    def ports(self) -> Dict[str, Port]:
        """This property allows you to access myref.ports, and receive a copy.

        of the ports dict which is correctly rotated and translated.
        """
        for name, port in self.parent.ports.items():
            port = self.parent.ports[name]
            new_center, new_orientation = self._transform_port(
                port.center,
                port.orientation,
                self.origin,
                self.rotation,
                self.x_reflection,
            )
            if name not in self._local_ports:
                self._local_ports[name] = port.copy(new_uid=True)
            self._local_ports[name].center = new_center
            self._local_ports[name].orientation = (
                mod(new_orientation, 360) if new_orientation else new_orientation
            )
            self._local_ports[name].parent = self
        # Remove any ports that no longer exist in the reference's parent
        parent_names = self.parent.ports.keys()
        local_names = list(self._local_ports.keys())
        for name in local_names:
            if name not in parent_names:
                self._local_ports.pop(name)
        return self._local_ports

    @property
    def info(self) -> Dict[str, Any]:
        return self.parent.info

    @property
    def metadata_child(self) -> Dict[str, Any]:
        return self.parent.metadata_child

    @property
    def size_info(self) -> SizeInfo:
        return SizeInfo(self.bbox)

    def pprint_ports(self) -> None:
        """Pretty print component ports."""
        ports_list = self.get_ports_list()
        for port in ports_list:
            print(port)

    def _transform_port(
        self,
        point: ndarray,
        orientation: float,
        origin: Coordinate = (0, 0),
        rotation: Optional[int] = None,
        x_reflection: bool = False,
    ) -> Tuple[ndarray, float]:
        """Apply GDS-type transformation to a port (x_ref)."""
        new_point = np.array(point)
        new_orientation = orientation

        if orientation is None:
            if origin is not None:
                new_point = new_point + np.array(origin)
            if x_reflection:
                new_point[1] = -new_point[1]
            return new_point, new_orientation

        if x_reflection:
            new_point[1] = -new_point[1]
            new_orientation = -orientation
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
            new_orientation += rotation
        if origin is not None:
            new_point = new_point + np.array(origin)
        new_orientation = mod(new_orientation, 360)

        return new_point, new_orientation

    def _transform_point(
        self,
        point: ndarray,
        origin: Coordinate = (0, 0),
        rotation: Optional[int] = None,
        x_reflection: bool = False,
    ) -> ndarray:
        """Apply GDS-type transformation to a point."""
        new_point = np.array(point)

        if x_reflection:
            new_point[1] = -new_point[1]
        if rotation is not None:
            new_point = _rotate_points(new_point, angle=rotation, center=[0, 0])
        if origin is not None:
            new_point = new_point + np.array(origin)

        return new_point

    def move(
        self,
        origin: Union[Port, Coordinate, str] = (0, 0),
        destination: Optional[Union[Port, Coordinate, str]] = None,
        axis: Optional[str] = None,
    ) -> "ComponentReference":
        """Move the ComponentReference from the origin point to the.

        destination.

        Both origin and destination can be 1x2 array-like, Port, or a key
        corresponding to one of the Ports in this device_ref.

        Args:
            origin: Port, port_name or Coordinate.
            destination: Port, port_name or Coordinate.
            axis: for the movemenent.

        Returns:
            ComponentReference.
        """
        # If only one set of coordinates is defined, make sure it's used to move things
        if destination is None:
            destination = origin
            origin = (0, 0)

        if isinstance(origin, str):
            if origin not in self.ports:
                raise ValueError(f"{origin} not in {self.ports.keys()}")

            origin = self.ports[origin]
            origin = cast(Port, origin)
            o = origin.center
        elif hasattr(origin, "center"):
            origin = cast(Port, origin)
            o = origin.center
        elif np.array(origin).size == 2:
            o = origin
        else:
            raise ValueError(
                f"move(origin={origin})\n"
                f"Invalid origin = {origin!r} needs to be"
                f"a coordinate, port or port name {list(self.ports.keys())}"
            )

        if isinstance(destination, str):
            if destination not in self.ports:
                raise ValueError(f"{destination} not in {self.ports.keys()}")

            destination = self.ports[destination]
            destination = cast(Port, destination)
            d = destination.center
        if hasattr(destination, "center"):
            destination = cast(Port, destination)
            d = destination.center
        elif np.array(destination).size == 2:
            d = destination

        else:
            raise ValueError(
                f"{self.parent.name}.move(destination={destination}) \n"
                f"Invalid destination = {destination!r} needs to be"
                f"a coordinate, a port, or a valid port name {list(self.ports.keys())}"
            )

        # Lock one axis if necessary
        if axis == "x":
            d = (d[0], o[1])
        if axis == "y":
            d = (o[0], d[1])

        # This needs to be done in two steps otherwise floating point errors can accrue
        dxdy = np.array(d) - np.array(o)
        self.origin = np.array(self.origin) + dxdy
        self._bb_valid = False
        return self

    def rotate(
        self,
        angle: float = 45,
        center: Union[Coordinate, str, int] = (0.0, 0.0),
    ) -> "ComponentReference":
        """Return rotated ComponentReference.

        Args:
            angle: in degrees
            center: x, y
        """
        if angle == 0:
            return self
        if isinstance(center, (int, str)):
            center = self.ports[center].center

        if isinstance(center, Port):
            center = center.center
        self.rotation += angle
        self.rotation %= 360
        self.origin = _rotate_points(self.origin, angle, center)
        self._bb_valid = False
        return self

    def reflect_h(
        self, port_name: Optional[str] = None, x0: Optional[Coordinate] = None
    ) -> "ComponentReference":
        """Perform horizontal mirror using x0 or port as axis (default, x0=0).

        This is the default for mirror along X=x0 axis
        """
        if port_name is None and x0 is None:
            x0 = -self.x

        if port_name is not None:
            position = self.ports[port_name]
            x0 = position.x
        self.reflect((x0, 1), (x0, 0))
        return self

    def reflect_v(
        self, port_name: Optional[str] = None, y0: Optional[float] = None
    ) -> "ComponentReference":
        """Perform vertical mirror using y0 as axis (default, y0=0)."""
        if port_name is None and y0 is None:
            y0 = 0.0

        if port_name is not None:
            position = self.ports[port_name]
            y0 = position.y
        self.reflect((1, y0), (0, y0))
        return self

    def mirror(
        self,
        p1: Coordinate = (0.0, 1.0),
        p2: Coordinate = (0.0, 0.0),
    ) -> "ComponentReference":
        """Mirrors.

        Args:
            p1: point 1.
            p2: point 2.
        """
        if isinstance(p1, Port):
            p1 = p1.center
        if isinstance(p2, Port):
            p2 = p2.center
        p1 = np.array(p1)
        p2 = np.array(p2)
        # Translate so reflection axis passes through origin
        self.origin = self.origin - p1

        # Rotate so reflection axis aligns with x-axis
        angle = np.arctan2((p2[1] - p1[1]), (p2[0] - p1[0])) * 180 / pi
        self.origin = _rotate_points(self.origin, angle=-angle, center=(0, 0))
        self.rotation -= angle

        # Reflect across x-axis
        self.x_reflection = not self.x_reflection
        self.origin[1] = -1 * self.origin[1]
        self.rotation = -1 * self.rotation

        # Un-rotate and un-translate
        self.origin = _rotate_points(self.origin, angle=angle, center=(0, 0))
        self.rotation += angle
        self.rotation = self.rotation % 360
        self.origin = self.origin + p1

        self._bb_valid = False
        return self

    def reflect(self, *args, **kwargs):
        warnings.warn(
            "reflect is deprecated and may be removed in a future version of gdsfactory. Use mirror instead.",
            DeprecationWarning,
        )
        return self.mirror(*args, **kwargs)

    def connect(
        self,
        port: Union[str, Port],
        destination: Port,
        overlap: float = 0.0,
    ) -> "ComponentReference":
        """Return ComponentReference where port connects to a destination.

        Args:
            port: origin (port or port_name) to connect.
            destination: destination port.
            overlap: how deep does the port go inside.

        Returns:
            ComponentReference: with correct rotation to connect to destination.
        """
        # port can either be a string with the name or an actual Port
        if port in self.ports:  # Then ``port`` is a key for the ports dict
            p = self.ports[port]
        elif isinstance(port, Port):
            p = port
        else:
            ports = list(self.ports.keys())
            raise ValueError(
                f"port = {port!r} not in {self.parent.name!r} ports {ports}"
            )

        angle = 180 + destination.orientation - p.orientation
        angle = angle % 360

        self.rotate(angle=angle, center=p.center)

        self.move(origin=p, destination=destination)
        self.move(
            -overlap
            * np.array(
                [
                    cos(destination.orientation * pi / 180),
                    sin(destination.orientation * pi / 180),
                ]
            )
        )

        return self

    def get_ports_list(self, **kwargs) -> List[Port]:
        """Return a list of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return list(select_ports(self.ports, **kwargs).values())

    def get_ports_dict(self, **kwargs) -> Dict[str, Port]:
        """Return a dict of ports.

        Keyword Args:
            layer: port GDS layer.
            prefix: port name prefix.
            orientation: in degrees.
            width: port width.
            layers_excluded: List of layers to exclude.
            port_type: optical, electrical, ...
            clockwise: if True, sort ports clockwise, False: counter-clockwise.
        """
        return select_ports(self.ports, **kwargs)

    @property
    def ports_layer(self) -> Dict[str, str]:
        """Return a mapping from layer0_layer1_E0: portName."""
        return map_ports_layer_to_orientation(self.ports)

    def port_by_orientation_cw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_cw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def port_by_orientation_ccw(self, key: str, **kwargs):
        """Return port by indexing them clockwise."""
        m = map_ports_to_orientation_ccw(self.ports, **kwargs)
        if key not in m:
            raise KeyError(f"{key} not in {list(m.keys())}")
        key2 = m[key]
        return self.ports[key2]

    def snap_ports_to_grid(self, nm: int = 1) -> None:
        for port in self.ports.values():
            port.snap_to_grid(nm=nm)

    def get_ports_xsize(self, **kwargs) -> float:
        """Return xdistance from east to west ports.

        Keyword Args:
            kwargs: orientation, port_type, layer.
        """
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return snap_to_grid(ports_ccw[0].x - ports_cw[0].x)

    def get_ports_ysize(self, **kwargs) -> float:
        """Returns ydistance from east to west ports."""
        ports_cw = self.get_ports_list(clockwise=True, **kwargs)
        ports_ccw = self.get_ports_list(clockwise=False, **kwargs)
        return snap_to_grid(ports_ccw[0].y - ports_cw[0].y)


def test_move():
    import gdsfactory as gf

    c = gf.Component()
    mzi = c.add_ref(gf.components.mzi())
    bend = c.add_ref(gf.components.bend_euler())
    bend.move("o1", mzi.ports["o2"])


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component()
    mzi = c.add_ref(gf.components.mzi())
    bend = c.add_ref(gf.components.bend_euler())
    bend.move("o1", mzi.ports["o2"])
    bend.move("o1", "o2")
    # c.show()



"""You can remove a list of layers from a component."""


import gdsfactory as gf
from gdsfactory.component import Component


@gf.cell
def test_remove_layers() -> Component:
    c = gf.Component("test_remove_layers")

    c.add_ref(gf.components.rectangle(size=(10, 1), layer=gf.LAYER.WG))
    c.add_ref(gf.components.rectangle(size=(10, 2), layer=gf.LAYER.SLAB90))
    c.add_ref(gf.components.rectangle(size=(10, 3), layer=gf.LAYER.SLAB150))

    assert len(c.layers) == 3

    c = c.remove_layers(layers=[gf.LAYER.SLAB90, gf.LAYER.SLAB150])

    assert len(c.layers) == 1
    return c


if __name__ == "__main__":
    c = test_remove_layers()
    c.show(show_ports=True)



"""FileWatcher based on watchdog. Looks for changes in files with .pic.yml extension."""

import logging
import pathlib
import sys
import time
import traceback
from functools import partial
from typing import Optional

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer

from gdsfactory.config import cwd
from gdsfactory.pdk import get_active_pdk
from gdsfactory.read.from_yaml import from_yaml


class YamlEventHandler(FileSystemEventHandler):
    """Captures pic.yml file change events."""

    def __init__(self, logger=None, path: Optional[str] = None):
        """Initialize the YAML event handler."""
        super().__init__()

        self.logger = logger or logging.root
        pdk = get_active_pdk()
        pdk.register_cells_yaml(dirpath=path)

    def update_cell(self, src_path, update: bool = False) -> None:
        """Register new YAML file into active pdk.

        pdk.cells[filename] = partial(from_yaml, filepath)

        """
        pdk = get_active_pdk()
        filepath = pathlib.Path(src_path)
        cell_name = filepath.stem.split(".")[0]
        function = partial(from_yaml, filepath)
        try:
            pdk.register_cells_yaml(**{cell_name: function}, update=update)
        except ValueError as e:
            print(e)

    def on_moved(self, event):
        super().on_moved(event)

        what = "directory" if event.is_directory else "file"
        if what == "file" and event.dest_path.endswith(".pic.yml"):
            self.logger.info("Moved %s: %s", what, event.src_path)
            self.update_cell(event.dest_path)
            self.get_component(event.src_path)

    def on_created(self, event):
        super().on_created(event)

        what = "directory" if event.is_directory else "file"
        if what == "file" and event.src_path.endswith(".pic.yml"):
            self.logger.info("Created %s: %s", what, event.src_path)
            self.update_cell(event.src_path)
            self.get_component(event.src_path)

    def on_deleted(self, event):
        super().on_deleted(event)

        what = "directory" if event.is_directory else "file"

        if what == "file" and event.src_path.endswith(".pic.yml"):
            self.logger.info("Deleted %s: %s", what, event.src_path)
            pdk = get_active_pdk()
            filepath = pathlib.Path(event.src_path)
            cell_name = filepath.stem.split(".")[0]
            pdk.remove_cell(cell_name)

    def on_modified(self, event):
        super().on_modified(event)

        what = "directory" if event.is_directory else "file"
        if (
            what == "file"
            and event.src_path.endswith(".pic.yml")
            or event.src_path.endswith(".py")
        ):
            self.logger.info("Modified %s: %s", what, event.src_path)
            self.get_component(event.src_path)

    def get_component(self, filepath):
        try:
            filepath = pathlib.Path(filepath)
            if filepath.exists():
                if str(filepath).endswith(".pic.yml"):
                    c = from_yaml(filepath)
                    self.update_cell(filepath, update=True)
                    c.show(show_ports=True)
                    # on_yaml_cell_modified.fire(c)
                    return c
                elif str(filepath).endswith(".py"):
                    d = dict(locals(), **globals())
                    d.update(__name__="__main__")
                    exec(filepath.read_text(), d, d)
                else:
                    print("Changed file {filepath} ignored (not .pic.yml or .py)")
        except Exception as e:
            traceback.print_exc(file=sys.stdout)
            print(e)


def watch(path=cwd) -> None:
    path = str(path)
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    event_handler = YamlEventHandler(path=path)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()
    logging.info(f"Observing {path!r}")
    try:
        while True:
            time.sleep(1)
    finally:
        observer.stop()
        observer.join()


if __name__ == "__main__":
    path = sys.argv[1] if len(sys.argv) > 1 else "."
    watch(path)



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.difftest import difftest


def test_get_route(data_regression: DataRegressionFixture, check: bool = True) -> None:
    c = gf.Component("sample_connect")
    mmi1 = c << gf.components.mmi1x2()
    mmi2 = c << gf.components.mmi1x2()
    mmi2.move((100, 50))
    route = gf.routing.get_route(
        mmi1.ports["o3"],
        mmi2.ports["o1"],
        cross_section=gf.cross_section.metal1,
        bend=gf.components.wire_corner,
    )
    c.add(route.references)
    if check:
        lengths = dict(length=route.length)

        data_regression.check(lengths)
        difftest(c)


def test_get_route_electrical_multilayer(
    data_regression: DataRegressionFixture, check: bool = True
) -> None:
    c = gf.Component("multi-layer")
    ptop = c << gf.components.pad_array()
    pbot = c << gf.components.pad_array(orientation=90)

    ptop.movex(300)
    ptop.movey(300)
    route = gf.routing.get_route_electrical_multilayer(
        ptop.ports["e11"],
        pbot.ports["e11"],
        end_straight_length=100,
    )
    c.add(route.references)
    if check:
        lengths = dict(length=route.length)

        data_regression.check(lengths)
        difftest(c)


if __name__ == "__main__":
    # c = gf.Component("sample_connect")
    # mmi1 = c << gf.components.mmi1x2()
    # mmi2 = c << gf.components.mmi1x2()
    # mmi2.move((100, 50))
    # route = gf.routing.get_route(
    #     mmi1.ports["o3"],
    #     mmi2.ports["o1"],
    #     cross_section=gf.cross_section.metal1,
    #     bend=gf.components.wire_corner,
    # )
    # c.add(route.references)
    # c.show(show_ports=True)

    c = gf.Component("multi-layer")
    ptop = c << gf.components.pad_array()
    pbot = c << gf.components.pad_array(orientation=90)

    ptop.movex(300)
    ptop.movey(300)
    route = gf.routing.get_route_electrical_multilayer(
        ptop.ports["e11"],
        pbot.ports["e11"],
        end_straight_length=100,
    )
    c.add(route.references)
    lengths = dict(length=route.length)
    c.show(show_ports=True)



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.component import Component


def test_get_bundle_electrical_multilayer(
    data_regression: DataRegressionFixture, check: bool = True
) -> Component:

    lengths = {}
    c = gf.Component("multi-layer")
    columns = 2
    ptop = c << gf.components.pad_array(columns=columns)
    pbot = c << gf.components.pad_array(orientation=90, columns=columns)

    ptop.movex(300)
    ptop.movey(300)
    routes = gf.routing.get_bundle_electrical_multilayer(
        ptop.ports, pbot.ports, end_straight_length=100, separation=20
    )

    for i, route in enumerate(routes):
        c.add(route.references)
        lengths[i] = route.length

    if check:
        data_regression.check(lengths)
    return c


if __name__ == "__main__":
    c = test_get_bundle_electrical_multilayer(None, check=False)
    c.show(show_ports=True)



"""You can also connect a component with single fiber INPUT and OUTPUTS (no fiber array)."""

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.samples.big_device import big_device


def test_fiber_single() -> Component:
    w = h = 18 * 50
    c = big_device(spacing=50.0, size=(w, h))
    return gf.routing.add_fiber_single(component=c)


if __name__ == "__main__":
    c = test_fiber_single()
    c.show(show_ports=True)



import gdsfactory as gf

if __name__ == "__main__":
    c = (gf.components.mzi_phase_shifter_top_heater_metal,)
    c = gf.routing.add_fiber_array(c)
    c = gf.routing.add_electrical_pads_top(c)
    c.show()

    # scene = c.to_3d()
    # scene.show()



import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.components.extension import extend_ports
from gdsfactory.components.taper import taper
from gdsfactory.components.text import text_rectangular
from gdsfactory.types import ComponentSpec, CrossSectionSpec, Float2, Optional

edge_coupler_silicon = gf.partial(taper, width2=0.2, length=100, with_two_ports=False)


@gf.cell
def edge_coupler_array(
    edge_coupler: ComponentSpec = edge_coupler_silicon,
    n: int = 5,
    pitch: float = 127.0,
    h_mirror: bool = False,
    v_mirror: bool = False,
    text: Optional[ComponentSpec] = text_rectangular,
    text_offset: Float2 = (10, 20),
) -> Component:
    """Fiber array edge coupler based on an inverse taper.

    Each edge coupler adds a ruler for polishing.

    Args:
        edge_coupler: edge coupler spec.
        n: number of channels.
        pitch: Fiber pitch.
        h_mirror: horizontal mirror.
        v_mirror: vertical mirror.
        text: text spec.
        text_offset: from edge coupler.
    """
    edge_coupler = gf.get_component(edge_coupler)

    c = Component()
    for i in range(n):
        ref = edge_coupler.ref(h_mirror=h_mirror, v_mirror=v_mirror)
        c.aliases[i] = ref
        ref.y = i * pitch
        c.add(ref)
        for port in ref.get_ports_list():
            c.add_port(f"{port.name}_{i}", port=port)

        if text:
            t = c << gf.get_component(text, text=str(i + 1))
            t.move(np.array(text_offset) + (0, i * pitch))

    c.auto_rename_ports()
    return c


@gf.cell
def edge_coupler_array_with_loopback(
    edge_coupler: ComponentSpec = edge_coupler_silicon,
    cross_section: CrossSectionSpec = "strip",
    radius: float = 30,
    n: int = 8,
    pitch: float = 127.0,
    extension_length: float = 1.0,
    h_mirror: bool = False,
    v_mirror: bool = False,
    right_loopback: bool = True,
    text: Optional[ComponentSpec] = text_rectangular,
    text_offset: Float2 = (0, 0),
) -> Component:
    """Fiber array edge coupler.

    Args:
        edge_coupler: edge coupler.
        cross_section: spec.
        radius: bend radius loopback (um).
        n: number of channels.
        pitch: Fiber pitch (um).
        extension_length: in um.
        h_mirror: horizontal mirror.
        v_mirror: vertical mirror.
        right_loopback: adds right loopback.
        text: Optional text spec.
        text_offset: x, y.
    """
    c = Component()
    ec = edge_coupler_array(
        edge_coupler=edge_coupler,
        n=n,
        pitch=pitch,
        h_mirror=h_mirror,
        v_mirror=v_mirror,
        text=text,
        text_offset=text_offset,
    )
    if extension_length > 0:
        ec = extend_ports(
            component=ec,
            port_names=("o1", "o2"),
            length=extension_length,
            extension=gf.partial(
                gf.c.straight, cross_section=cross_section, length=extension_length
            ),
        )

    ec_ref = c << ec
    route1 = gf.routing.get_route(
        ec_ref.ports["o1"],
        ec_ref.ports["o2"],
        cross_section=cross_section,
        radius=radius,
    )
    c.add(route1.references)

    if n > 4 and right_loopback:
        route2 = gf.routing.get_route(
            ec_ref.ports[f"o{n-1}"],
            ec_ref.ports[f"o{n}"],
            cross_section=cross_section,
            radius=radius,
        )
        c.add(route2.references)
        for i in range(n - 4):
            c.add_port(str(i), port=ec_ref.ports[f"o{i+3}"])
    elif n > 4:
        for i in range(n - 2):
            c.add_port(str(i), port=ec_ref.ports[f"o{i+3}"])

    c.auto_rename_ports()
    return c


if __name__ == "__main__":
    # c = edge_coupler_silicon()
    # c = edge_coupler_array()
    c = edge_coupler_array_with_loopback()
    c.show(show_ports=True)



import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.types import Float2, Layer

layer = gf.LAYER.WG


@gf.cell
def width_min(size: Float2 = (0.1, 0.1)) -> Component:
    return gf.components.rectangle(size=size, layer=layer)


@gf.cell
def area_min() -> Component:
    size = (0.2, 0.2)
    return gf.components.rectangle(size=size, layer=layer)


@gf.cell
def gap_min(gap: float = 0.1) -> Component:
    c = gf.Component()
    r1 = c << gf.components.rectangle(size=(1, 1), layer=layer)
    r2 = c << gf.components.rectangle(size=(1, 1), layer=layer)
    r1.xmax = 0
    r2.xmin = gap
    return c


@gf.cell
def separation(
    gap: float = 0.1, layer1: Layer = gf.LAYER.HEATER, layer2: Layer = gf.LAYER.M1
) -> Component:
    c = gf.Component()
    r1 = c << gf.components.rectangle(size=(1, 1), layer=layer1)
    r2 = c << gf.components.rectangle(size=(1, 1), layer=layer2)
    r1.xmax = 0
    r2.xmin = gap
    return c


@gf.cell
def enclosing(
    enclosing: float = 0.1, layer1: Layer = gf.LAYER.VIAC, layer2: Layer = gf.LAYER.M1
) -> Component:
    """Layer1 must be enclosed by layer2 by value.

    checks if layer1 encloses (is bigger than) layer2 by value
    """
    w1 = 1
    w2 = w1 + enclosing
    c = gf.Component()
    c << gf.components.rectangle(size=(w1, w1), layer=layer1, centered=True)
    r2 = c << gf.components.rectangle(size=(w2, w2), layer=layer2, centered=True)
    r2.movex(0.5)
    return c


@gf.cell
def snapping_error(gap: float = 1e-3) -> Component:
    c = gf.Component()
    r1 = c << gf.components.rectangle(size=(1, 1), layer=layer)
    r2 = c << gf.components.rectangle(size=(1, 1), layer=layer)
    r1.xmax = 0
    r2.xmin = gap
    return c


@gf.cell
def errors() -> Component:
    components = [width_min(), gap_min(), separation(), enclosing()]
    c = gf.pack(components, spacing=1.5)
    return c[0]


if __name__ == "__main__":
    # c = width_min()
    # c.write_gds("wmin.gds")
    # c = gap_min()
    # c.write_gds("gmin.gds")
    # c = snapping_error()
    # c.write_gds("snap.gds")

    c = errors()
    c = gf.add_padding_container(c, layers=(gf.LAYER.FLOORPLAN,), default=5)
    c.write_gds("errors.gds")
    c.show(show_ports=True)



from typing import Optional, Tuple

import numpy as np

import gdsfactory as gf
from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.ge_detector_straight_si_contacts import (
    ge_detector_straight_si_contacts,
)
from gdsfactory.components.mmi_90degree_hybrid import mmi_90degree_hybrid
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.types import ComponentSpec, CrossSectionSpec


@cell
def coh_rx_single_pol(
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight_function,
    cross_section: CrossSectionSpec = "strip",
    hybrid_90deg: ComponentSpec = mmi_90degree_hybrid,
    detector: ComponentSpec = ge_detector_straight_si_contacts,
    det_spacing: Tuple[float, float] = (60.0, 50.0),
    with_pads: bool = True,
    pad_det_spacing: float = 80.0,
    in_wg_length: float = 20.0,
    lo_input_coupler: Optional[ComponentSpec] = None,
    signal_input_coupler: Optional[ComponentSpec] = None,
) -> Component:
    r"""Single polarization coherent receiver.

    Args:
        bend: 90 degrees bend library.
        straight: straight function.
        cross_section: for routing.
        hybrid_90deg: generates the 90 degree hybrid.
        detector: generates the detector.
        det_spacing: spacing between 90 degree hybrid and detector and
           vertical spacing between detectors.
        with_pads: if True, it draws pads for the balanced detectors.
        pad_det_spacing: spacing between the pads and the detectors (if with_pads=True).
        in_wg_length: length of the straight waveguides at the input of the 90 deg hybrid.
        lo_input_coupler: Optional coupler for the LO.
        signal_input_coupler: Optional coupler for the signal.

    .. code::

                               _________
           (lo_in_coupler)---|          |--- detI1 \\ __ i signal
                             |  90 deg  |--- detI2 //
       (signal_in_coupler)---|  hybrid  |--- detQ1 \\ __ q signal
                             |__________|--- detQ2 //
    """
    bend = gf.get_component(bend, cross_section=cross_section)
    straight_spec = straight

    # ----- Draw 90 deg hybrid -----

    c = Component()

    hybrid_90deg = gf.get_component(hybrid_90deg)
    hybrid = c << hybrid_90deg

    # ----- Draw input waveguides (and coupler if indicated) ---

    if in_wg_length > 0.0:
        straight = gf.get_component(straight_spec, length=in_wg_length)
        signal_in = c << straight
        lo_in = c << straight

        signal_in.connect("o2", hybrid.ports["signal_in"])
        lo_in.connect("o2", hybrid.ports["LO_in"])

    if lo_input_coupler is not None:
        lo_in_coupler = gf.get_component(lo_input_coupler)
        in_coup_lo = c << lo_in_coupler

        if in_wg_length > 0.0:
            in_coup_lo.connect("o1", lo_in.ports["o1"])
        else:
            in_coup_lo.connect("o1", hybrid.ports["LO_in"])

    elif in_wg_length > 0.0:
        c.add_port("LO_in", port=lo_in.ports["o1"])
    else:
        c.add_port("LO_in", port=hybrid.ports["LO_in"])

    if signal_input_coupler is not None:
        signal_in_coupler = gf.get_component(signal_input_coupler)
        in_coup_signal = c << signal_in_coupler

        if in_wg_length > 0.0:
            in_coup_signal.connect("o1", signal_in.ports["o1"])
        else:
            in_coup_signal.connect("o1", hybrid.ports["signal_in"])

    elif in_wg_length > 0.0:
        c.add_port("signal_in", port=signal_in.ports["o1"])
    else:
        c.add_port("signal_in", port=hybrid.ports["signal_in"])

    # ---- Draw photodetectors -----

    detector = gf.get_component(detector)

    pd_i1 = c << detector
    pd_i2 = c << detector
    pd_q1 = c << detector
    pd_q2 = c << detector

    pds = [pd_i1, pd_i2, pd_q1, pd_q2]

    # x placement
    for pd in pds:
        pd.xmin = hybrid.xmax + det_spacing[0]

    # y placement - we will place them in the same order as the outputs
    # of the 90 degree hybrid to avoid crossings
    hybrid_ports = {"I_out1": pd_i1, "I_out2": pd_i2, "Q_out1": pd_q1, "Q_out2": pd_q2}

    port_names = hybrid_ports.keys()
    ports_y_pos = [hybrid.ports[port_name].y for port_name in port_names]
    inds = np.argsort(ports_y_pos)
    port_names = list(port_names)
    port_names = [port_names[i] for i in inds]

    y_pos = hybrid.y - 1.5 * det_spacing[1]

    det_ports = []
    ports_hybrid = []
    for port_name in port_names:
        det = hybrid_ports[port_name]
        det.y = y_pos
        y_pos = y_pos + det_spacing[1]
        det_ports.append(det.ports["o1"])
        ports_hybrid.append(hybrid.ports[port_name])

    routes = gf.routing.get_bundle(ports_hybrid, det_ports)
    for route in routes:
        c.add(route.references)

    # --- Draw metal connections ----

    route = gf.routing.get_route_electrical(
        pd_i1.ports["bot_e3"], pd_i2.ports["top_e3"]
    )
    c.add(route.references)

    # Add a port at the center
    x_max = -np.inf
    for ref in route.references:
        if ref.xmax > x_max:
            x_max = ref.xmax
    c.add_port(
        name="i_out",
        port_type="placement",
        layer="M3",
        center=(x_max, (pd_i1.ports["bot_e3"].y + pd_i2.ports["top_e3"].y) / 2),
        orientation=0,
        width=2.0,
    )

    route = gf.routing.get_route_electrical_m2(
        pd_q1.ports["bot_e3"], pd_q2.ports["top_e3"]
    )
    c.add(route.references)

    # Add a port
    x_max = -np.inf
    for ref in route.references:
        if ref.xmax > x_max:
            x_max = ref.xmax
    c.add_port(
        name="q_out",
        port_type="placement",
        layer="M2",
        center=(
            x_max,
            (pd_q1.ports["bot_e3"].y + pd_q2.ports["top_e3"].y) / 2 - 15.0,
        ),  # - 20.0 so that the traces for I and Q do not overlap
        orientation=0,
        width=2.0,
    )

    # --- Draw pads if indicated ----

    if with_pads:

        pad_array = c << gf.components.pad_array(columns=1, rows=4, orientation=0)
        pad_array.xmin = pd_i1.xmax + pad_det_spacing
        pad_array.y = hybrid.y

        # Add labels
        labels = {"e11": "V+", "e21": "Q_out", "e31": "I_out", "e41": "V-"}
        for pad, label in labels.items():
            x_pos, y_pos = pad_array.ports[pad].center
            c << gf.components.text(
                text=label,
                size=14.0,
                position=[x_pos + 55.0, y_pos],
                justify="left",
                layer="M3",
            )

        # Connect to the pads. Need to do it manually to avoid crossings

        # V- pad (connected to positive side of one of the diodes)
        p0x, p0y = pd_i1.ports["top_e2"].center
        p1x, p1y = pad_array.ports["e41"].center
        route = gf.routing.get_route_from_waypoints_electrical(
            [(p0x, p0y), (p0x, p1y), (p1x, p1y)]
        )

        c.add(route.references)

        p0x, p0y = pd_q1.ports["top_e3"].center
        p1x, p1y = pad_array.ports["e41"].center
        route = gf.routing.get_route_from_waypoints_electrical_multilayer(
            [
                (p0x, p0y),
                (p0x + 0.5 * (p1x - p0x), p0y),
                (p0x + 0.5 * (p1x - p0x), p1y),
            ]
        )

        c.add(route.references)

        # V+ pad (connected to negative side of the other diode)
        p0x, p0y = pd_i2.ports["bot_e2"].center
        p1x, p1y = pad_array.ports["e11"].center
        route = gf.routing.get_route_from_waypoints_electrical(
            [(p0x, p0y), (p0x, p1y), (p1x, p1y)]
        )
        c.add(route.references)

        p0x, p0y = pd_q2.ports["bot_e3"].center
        p1x, p1y = pad_array.ports["e11"].center
        route = gf.routing.get_route_from_waypoints_electrical_multilayer(
            [
                (p0x, p0y),
                (p0x + 0.5 * (p1x - p0x), p0y),
                (p0x + 0.5 * (p1x - p0x), p1y),
            ]
        )

        c.add(route.references)

        # I out pad
        route = gf.routing.get_route_electrical(
            pad_array.ports["e31"], c.ports["i_out"]
        )
        c.add(route.references)

        # Q out pad
        route = gf.routing.get_route_electrical_multilayer(
            pad_array.ports["e21"], c.ports["q_out"]
        )
        c.add(route.references)

    else:
        # Create electrical ports. q_out and i_out already exist
        c.add_ports(
            pd_i1.get_ports_list(port_type="electrical", prefix="top"),
            prefix="i1vminus",
        )
        c.add_ports(
            pd_q1.get_ports_list(port_type="electrical", prefix="top"),
            prefix="q1vminus",
        )
        c.add_ports(
            pd_q2.get_ports_list(port_type="electrical", prefix="bot"), prefix="q2vplus"
        )
        c.add_ports(
            pd_i2.get_ports_list(port_type="electrical", prefix="bot"), prefix="i2vplus"
        )

    return c


if __name__ == "__main__":
    c = coh_rx_single_pol(with_pads=False)
    c.show(show_ports=True)



import os
import warnings

import numpy as np

import gdsfactory as gf
from gdsfactory.component import Component
from gdsfactory.constants import _glyph, _indent, _width
from gdsfactory.types import LayerSpec


@gf.cell
def text_freetype(
    text: str = "abcd",
    size: int = 10,
    justify: str = "left",
    layer: LayerSpec = "WG",
    font: str = "DEPLOF",
) -> Component:
    """Returns text Component.

    Args:
        text: string.
        size: in um.
        position: x, y position.
        justify: left, right, center.
        layer: for the text.
        font: Font face to use. Default DEPLOF does not require additional libraries,
            otherwise freetype load fonts. You can choose font by name
            (e.g. "Times New Roman"), or by file OTF or TTF filepath.
    """
    t = Component()
    xoffset = 0
    yoffset = 0

    face = font
    if face == "DEPLOF":
        scaling = size / 1000

        for line in text.split("\n"):
            char = Component()
            for c in line:
                ascii_val = ord(c)
                if c == " ":
                    xoffset += 500 * scaling
                elif (33 <= ascii_val <= 126) or (ascii_val == 181):
                    for poly in _glyph[ascii_val]:
                        xpts = np.array(poly)[:, 0] * scaling
                        ypts = np.array(poly)[:, 1] * scaling
                        char.add_polygon([xpts + xoffset, ypts + yoffset], layer=layer)
                    xoffset += (_width[ascii_val] + _indent[ascii_val]) * scaling
                else:
                    valid_chars = "!\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~µ"
                    warnings.warn(
                        'text(): Warning, some characters ignored, no geometry for character "%s" with ascii value %s. '
                        "Valid characters: %s"
                        % (chr(ascii_val), ascii_val, valid_chars)
                    )
            t.add_ref(char)
            yoffset -= 1500 * scaling
            xoffset = 0
    else:
        from gdsfactory.font import _get_font_by_file, _get_font_by_name, _get_glyph

        # Load the font
        # If we've passed a valid file, try to load that, otherwise search system fonts
        font = None
        if (face.endswith(".otf") or face.endswith(".ttf")) and os.path.exists(face):
            font = _get_font_by_file(face)
        else:
            try:
                font = _get_font_by_name(face)
            except ValueError:
                pass
        if font is None:
            raise ValueError(
                f"Failed to find font: {face!r}. "
                "Try specifying the exact (full) path to the .ttf or .otf file. "
            )

        # Render each character
        for line in text.split("\n"):
            char = Component()
            xoffset = 0
            for letter in line:
                letter_dev = Component()
                letter_template, advance_x = _get_glyph(font, letter)
                for poly in letter_template.polygons:
                    letter_dev.add_polygon(poly.polygons, layer=layer)
                ref = char.add_ref(letter_dev)
                ref.move(destination=(xoffset, 0))
                ref.magnification = size
                xoffset += size * advance_x

            ref = t.add_ref(char)
            ref.move(destination=(0, yoffset))
            yoffset -= size
            t.absorb(ref)

    justify = justify.lower()
    for ref in t.references:
        if justify == "center":
            ref.move(origin=ref.center, destination=(0, 0), axis="x")

        elif justify == "right":
            ref.xmax = 0
    t.flatten()
    return t


if __name__ == "__main__":
    c2 = text_freetype(
        "hello",
        # font="Times New Roman"
    )
    c2.show(show_ports=True)



"""Store files."""

from io import BytesIO

import numpy as np
from pydantic import BaseModel
from typing_extensions import Literal

from gdsfactory.config import CONFIG
from gdsfactory.types import Optional, PathType

FileTypes = Literal["sparameters", "modes", "gds", "measurements"]


class FileStorage(BaseModel):
    dirpath: Optional[PathType] = CONFIG["gdslib"]
    filetype: FileTypes

    def write(self, filename: str, data):
        raise NotImplementedError("need to implement")

    def read(self, filename: str) -> np.ndarray:
        raise NotImplementedError("need to implement")

    def _write_local_cache(self, filename: str, data):
        if not self.dirpath:
            return
        filepath = self.dirpath / f"{self.filetype}/{filename}.npz"

        with open(filepath, "wb") as file:
            np.savez_compressed(file, **data)

    def _read_local_cache(self, filename: str) -> Optional[np.ndarray]:
        if not self.dirpath:
            return
        filepath = self.dirpath / f"{self.filetype}/{filename}.npz"
        if filepath.exists():
            return np.load(filepath)


class FileStorageGoogleCloud(FileStorage):
    bucket_name: str

    def write(self, filename: str, data) -> None:
        try:
            from google.cloud import storage
        except ImportError as e:
            raise ImportError("pip install google-cloud-storage") from e

        filepath = f"{self.filetype}/{filename}.npz"
        storage_client = storage.Client()
        bucket = storage_client.bucket(self.bucket_name)
        blob = bucket.blob(filepath)

        with blob.open("wb", ignore_flush=True) as file:
            np.savez_compressed(file, **data)

    def read(self, filename: str) -> np.ndarray:
        data = self._read_local_cache(filename=filename)
        if data:
            return data

        try:
            from google.cloud import storage
        except ImportError as e:
            raise ImportError("pip install google-cloud-storage") from e

        filepath = f"{self.filetype}/{filename}.npz"
        storage_client = storage.Client()
        bucket = storage_client.bucket(self.bucket_name)
        blob = bucket.blob(filepath)
        b = blob.download_as_bytes()
        return np.load(BytesIO(b))


# def test_google_cloud() -> None:
#     w = np.linspace(1.5, 1.6, 3)
#     s = dict(wavelengths=w)
#     s["o1@0,o2@0"] = np.zeros_like(w)

#     f = FileStorageGoogleCloud(bucket_name="gdsfactory", filetype="sparameters")
#     # f.write("demo", s)

#     field = "o1@0,o2@0"
#     s2 = f.read("demo")
#     np.isclose(s[field], s2[field])


if __name__ == "__main__":
    # test_google_cloud()

    w = np.linspace(1.5, 1.6, 3)
    s = dict(wavelengths=w)
    field = "o1@0,o2@0"
    s["o1@0,o2@0"] = np.zeros_like(w)

    f = FileStorageGoogleCloud(bucket_name="gdsfactory", filetype="sparameters")
    # f.write("demo", s)

    s2 = f.read("demo")
    print(s2["o1@0,o2@0"])



import gdsfactory as gf


def test_route_error2():
    """Ensures that an impossible route raises value Error."""
    c = gf.Component("pads_route_from_steps")
    pt = c << gf.components.pad_array(orientation=270, columns=3)
    pb = c << gf.components.pad_array(orientation=90, columns=3)
    pt.move((100, 200))
    route = gf.routing.get_route_from_steps(
        pt.ports["e11"],
        pb.ports["e11"],
        steps=[
            {"y": 100},
        ],
        cross_section="metal_routing",
        bend=gf.components.wire_corner,
    )
    c.add(route.references)
    return c


if __name__ == "__main__":
    c = test_route_error2()
    c.show(show_ports=True)



import numpy as np

from gdsfactory.cell import cell
from gdsfactory.component import Component
from gdsfactory.types import LayerSpec


@cell
def optimal_step(
    start_width: float = 10,
    end_width: float = 22,
    num_pts: int = 50,
    width_tol: float = 1e-3,
    anticrowding_factor: float = 1.2,
    symmetric: bool = False,
    layer: LayerSpec = (1, 0),
) -> Component:
    """Returns an optimally-rounded step geometry.

    Args:
        start_width: Width of the connector on the left end of the step.
        end_width: Width of the connector on the right end of the step.
        num_pts: number of points comprising the entire step geometry.
        width_tol: Point at which to terminate the calculation of the optimal step
        anticrowding_factor: Factor to reduce current crowding by elongating
            the structure and reducing the curvature
        symmetric: If True, adds a mirrored copy of the step across the x-axis to the
            geometry and adjusts the width of the ports.
        layer: layer spec to put polygon geometry on.

    based on phidl.geometry

    Notes
    -----
    Optimal structure from https://doi.org/10.1103/PhysRevB.84.174510
    Clem, J., & Berggren, K. (2011). Geometry-dependent critical currents in
    superconducting nanocircuits. Physical Review B, 84(17), 1–27.
    """
    # ==========================================================================
    #  Create the basic geometry
    # ==========================================================================
    def step_points(eta, W, a):
        # Returns points from a unit semicircle in the w (= u + iv) plane to
        # the optimal curve in the zeta (= x + iy) plane which transitions
        # a wire from a width of 'W' to a width of 'a'
        # eta takes value 0 to pi

        W = complex(W)
        a = complex(a)

        gamma = (a**2 + W**2) / (a**2 - W**2)

        w = np.exp(1j * eta)

        zeta = (
            4
            * 1j
            / np.pi
            * (
                W * np.arctan(np.sqrt((w - gamma) / (gamma + 1)))
                + a * np.arctan(np.sqrt((gamma - 1) / (w - gamma)))
            )
        )

        x = np.real(zeta)
        y = np.imag(zeta)
        return x, y

    def invert_step_point(x_desired=-10, y_desired=None, W=1, a=2):
        # Finds the eta associated with the value x_desired along the
        # optimal curve
        def fh(eta):
            guessed_x, guessed_y = step_points(eta, W=W, a=a)
            if y_desired is None:
                return (guessed_x - x_desired) ** 2  # The error
            else:
                return (guessed_y - y_desired) ** 2

        try:
            from scipy.optimize import fminbound
        except Exception as e:
            raise ImportError(
                "To run the optimal-curve geometry "
                "functions you need scipy, please install "
                "it with `pip install scipy`"
            ) from e
        found_eta = fminbound(fh, x1=0, x2=np.pi, args=())
        return step_points(found_eta, W=W, a=a)

    if start_width > end_width:
        reverse = True
        start_width, end_width = end_width, start_width
    else:
        reverse = False

    D = Component()
    if start_width == end_width:  # Just return a square
        if symmetric:
            ypts = [
                -start_width / 2,
                start_width / 2,
                start_width / 2,
                -start_width / 2,
            ]
            xpts = [0, 0, start_width, start_width]
        if not symmetric:
            ypts = [0, start_width, start_width, 0]
            xpts = [0, 0, start_width, start_width]
        D.info["num_squares"] = 1
    else:
        xmin, ymin = invert_step_point(
            y_desired=start_width * (1 + width_tol), W=start_width, a=end_width
        )
        xmax, ymax = invert_step_point(
            y_desired=end_width * (1 - width_tol), W=start_width, a=end_width
        )

        xpts = np.linspace(xmin, xmax, num_pts).tolist()
        ypts = []
        for x in xpts:
            x, y = invert_step_point(x_desired=x, W=start_width, a=end_width)
            ypts.append(y)

        ypts[-1] = end_width
        ypts[0] = start_width
        y_num_sq = np.array(ypts)
        x_num_sq = np.array(xpts)

        if not symmetric:
            xpts.append(xpts[-1])
            ypts.append(0)
            xpts.append(xpts[0])
            ypts.append(0)
        else:
            xpts += list(xpts[::-1])
            ypts += [-y for y in ypts[::-1]]
            xpts = [x / 2 for x in xpts]
            ypts = [y / 2 for y in ypts]

        # anticrowding_factor stretches the wire out; a stretched wire is a
        # gentler transition, so there's less chance of current crowding if
        # the fabrication isn't perfect but as a result, the wire isn't as
        # short as it could be
        xpts = (np.array(xpts) * anticrowding_factor).tolist()

        if reverse:
            xpts = (-np.array(xpts)).tolist()
            start_width, end_width = end_width, start_width

        D.info["num_squares"] = float(
            np.round(
                np.sum(np.diff(x_num_sq) / ((y_num_sq[:-1] + y_num_sq[1:]) / 2)), 3
            )
        )

    # ==========================================================================
    #  Create a blank device, add the geometry, and define the ports
    # ==========================================================================
    D.add_polygon([xpts, ypts], layer=layer)

    if not symmetric:
        D.add_port(
            name="e1",
            center=[min(xpts), start_width / 2],
            width=start_width,
            orientation=180,
            layer=layer,
        )
        D.add_port(
            name="e2",
            center=[max(xpts), end_width / 2],
            width=end_width,
            orientation=0,
            layer=layer,
        )
    if symmetric:
        D.add_port(
            name="e1",
            center=[min(xpts), 0],
            width=start_width,
            orientation=180,
            layer=layer,
        )
        D.add_port(
            name="e2",
            center=[max(xpts), 0],
            width=end_width,
            orientation=0,
            layer=layer,
        )

    return D


if __name__ == "__main__":
    c = optimal_step()
    print(c.to_dict())
    # c = optimal_step()
    # print(c.to_dict())
    c.show(show_ports=True)



import numpy as np


def ring(
    wl: np.ndarray,
    wl0: float,
    neff: float,
    ng: float,
    ring_length: float,
    coupling: float,
    loss: float,
) -> np.ndarray:
    """Returns Frequency Domain Response of an all pass filter.

    Args:
        wl: in um.
        wl0: the wavelength at which neff and ng are defined.
        neff: effective index.
        ng: group index.
        ring_length: in um.
        loss: dB/um.
    """
    transmission = 1 - coupling
    neff_wl = (
        neff + (wl0 - wl) * (ng - neff) / wl0
    )  # we expect a linear behavior with respect to wavelength
    out = np.sqrt(transmission) - 10 ** (-loss * ring_length / 20.0) * np.exp(
        2j * np.pi * neff_wl * ring_length / wl
    )
    out /= 1 - np.sqrt(transmission) * 10 ** (-loss * ring_length / 20.0) * np.exp(
        2j * np.pi * neff_wl * ring_length / wl
    )
    return abs(out) ** 2


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    loss = 0.03  # [dB/μm] (alpha) waveguide loss
    neff = 2.46  # Effective index of the waveguides
    wl0 = 1.55  # [μm] the wavelength at which neff and ng are defined
    radius = 5
    ring_length = 2 * np.pi * radius  # [μm] Length of the ring
    coupling = 0.5  # [] coupling of the coupler
    wl = np.linspace(1.5, 1.6, 1000)  # [μm] Wavelengths to sweep over
    wl = np.linspace(1.55, 1.60, 1000)  # [μm] Wavelengths to sweep over
    ngs = [4.182551, 4.169563, 4.172917]
    thicknesses = [210, 220, 230]

    # widths = np.array([0.4, 0.45, 0.5, 0.55, 0.6])
    # ngs = np.array([4.38215238, 4.27254985, 4.16956338, 4.13283219, 4.05791982])

    widths = np.array([0.495, 0.5, 0.505])
    neffs = np.array([2.40197253, 2.46586378, 2.46731758])
    ng = 4.2  # Group index of the waveguides

    for width, neff in zip(widths, neffs):
        p = ring(
            wl=wl,
            wl0=wl0,
            neff=neff,
            ng=ng,
            ring_length=ring_length,
            coupling=coupling,
            loss=loss,
        )
        plt.plot(wl, p, label=f"{int(width*1e3)}nm")

    # for thickness, ng in zip(thicknesses, ngs):
    #     p = ring(
    #         wl=wl,
    #         wl0=wl0,
    #         neff=neff,
    #         ng=ng,
    #         ring_length=ring_length,
    #         coupling=coupling,
    #         loss=loss,
    #     )
    #     plt.plot(wl, p, label=str(thickness))

    plt.title("ring resonator vs waveguide width")
    plt.xlabel("wavelength (um)")
    plt.ylabel("Power Transmission")
    plt.grid()
    plt.legend()
    plt.show()



from pytest_regressions.data_regression import DataRegressionFixture

import gdsfactory as gf
from gdsfactory.difftest import difftest


def test_import_from_phidl(data_regression: DataRegressionFixture) -> None:
    """Read ports from markers."""
    import phidl.geometry as pg

    c1 = pg.ytron_round()
    c2 = gf.read.from_phidl(component=c1)
    data_regression.check(c2.to_dict())
    difftest(c2)


if __name__ == "__main__":
    import phidl.geometry as pg

    c1 = pg.ytron_round()
    c2 = gf.read.from_phidl(component=c1)
    c2.show(show_ports=True)



from typing import Dict, List, Optional, Union

import numpy as np

import gdsfactory as gf
from gdsfactory.components.bend_euler import bend_euler
from gdsfactory.components.straight import straight as straight_function
from gdsfactory.components.taper import taper as taper_function
from gdsfactory.components.via_corner import via_corner
from gdsfactory.components.wire import wire_corner
from gdsfactory.port import Port
from gdsfactory.routing.get_bundle_from_waypoints import get_bundle_from_waypoints
from gdsfactory.routing.manhattan import _is_horizontal, _is_vertical
from gdsfactory.routing.sort_ports import sort_ports as sort_ports_function
from gdsfactory.types import (
    ComponentSpec,
    CrossSectionSpec,
    MultiCrossSectionAngleSpec,
    Route,
)


def get_bundle_from_steps(
    ports1: List[Port],
    ports2: List[Port],
    steps: Optional[List[Dict[str, float]]] = None,
    bend: ComponentSpec = bend_euler,
    straight: ComponentSpec = straight_function,
    taper: Optional[ComponentSpec] = taper_function,
    cross_section: Union[CrossSectionSpec, MultiCrossSectionAngleSpec] = "strip",
    sort_ports: bool = True,
    separation: Optional[float] = None,
    path_length_match_loops: int = None,
    path_length_match_extra_length: float = 0.0,
    path_length_match_modify_segment_i: int = -2,
    **kwargs
) -> List[Route]:
    """Returns a list of routes formed by the given waypoints steps.

    Can add bends instead of corners and optionally tapers in straight sections.
    Tapering to wider straights reduces the optical loss and phase errors.
    `get_bundle_from_steps` is a manual version of `get_bundle`
    and a more convenient version of `get_bundle_from_waypoints`.

    Args:
        port1: start ports (list or dict).
        port2: end ports (list or dict).
        steps: changes that define the route [{'dx': 5}, {'dy': 10}].
        bend: function that returns bends.
        straight: function that returns straight waveguides.
        taper: function that returns tapers.
        cross_section: for routes.
        sort_ports: if True sort ports.
        separation: center to center, defaults to ports1 separation.
        kwargs: cross_section settings.

    .. plot::
        :include-source:

        import gdsfactory as gf

        c = gf.Component("get_route_from_steps_sample")
        w = gf.components.array(
            gf.partial(gf.components.straight, layer=(2, 0)),
            rows=3,
            columns=1,
            spacing=(0, 50),
        )

        left = c << w
        right = c << w
        right.move((200, 100))
        p1 = left.get_ports_list(orientation=0)
        p2 = right.get_ports_list(orientation=180)

        routes = gf.routing.get_bundle_from_steps(
            p1,
            p2,
            steps=[{"x": 150}],
        )

        for route in routes:
            c.add(route.references)
        c.plot()
        c.show(show_ports=True)

    """
    if isinstance(ports1, Port):
        ports1 = [ports1]

    if isinstance(ports2, Port):
        ports2 = [ports2]

    # convert ports dict to list
    if isinstance(ports1, dict):
        ports1 = list(ports1.values())

    if isinstance(ports2, dict):
        ports2 = list(ports2.values())

    if sort_ports:
        ports1, ports2 = sort_ports_function(ports1, ports2)

    waypoints = []
    steps = steps or []

    x, y = ports1[0].center
    for d in steps:
        x = d["x"] if "x" in d else x
        x += d.get("dx", 0)
        y = d["y"] if "y" in d else y
        y += d.get("dy", 0)
        waypoints += [(x, y)]

    port2 = ports2[0]
    x2, y2 = port2.center
    orientation = port2.orientation
    if orientation is None:
        p1 = waypoints[-2]
        p0 = waypoints[-1]
        if _is_vertical(p0, p1):
            waypoints += [(y2, y)]
        elif _is_horizontal(p0, p1):
            waypoints += [(x, x2)]
    elif int(orientation) in {0, 180}:
        waypoints += [(x, y2)]
    elif int(orientation) in {90, 270}:
        waypoints += [(x2, y)]
    waypoints = np.array(waypoints)

    if not isinstance(cross_section, list):
        x = gf.get_cross_section(cross_section, **kwargs)
        auto_widen = x.auto_widen

        if auto_widen:
            taper = gf.get_component(
                taper,
                length=x.taper_length,
                width1=x.width,
                width2=x.width_wide,
                cross_section=cross_section,
                **kwargs,
            )
        else:
            taper = None
    else:
        taper = None

    return get_bundle_from_waypoints(
        ports1=ports1,
        ports2=ports2,
        waypoints=waypoints,
        bend=bend,
        straight=straight,
        taper=taper,
        cross_section=cross_section,
        separation=separation,
        path_length_match_extra_length=path_length_match_extra_length,
        path_length_match_modify_segment_i=path_length_match_modify_segment_i,
        path_length_match_loops=path_length_match_loops,
        **kwargs,
    )


get_bundle_from_steps_electrical = gf.partial(
    get_bundle_from_steps, bend=wire_corner, cross_section="metal_routing"
)

get_bundle_from_steps_electrical_multilayer = gf.partial(
    get_bundle_from_steps,
    bend=via_corner,
    cross_section=[
        (gf.cross_section.metal2, (90, 270)),
        ("metal_routing", (0, 180)),
    ],
)


def _demo() -> None:
    c = gf.Component("get_route_from_steps_sample")

    w = gf.components.array(
        gf.partial(gf.components.straight, layer=(2, 0)),
        rows=3,
        columns=1,
        spacing=(0, 50),
    )

    left = c << w
    right = c << w
    right.move((200, 100))
    p1 = left.get_ports_list(orientation=0)
    p2 = right.get_ports_list(orientation=180)

    routes = get_bundle_from_steps_electrical(
        p1,
        p2,
        steps=[{"x": 150}],
    )

    for route in routes:
        c.add(route.references)

    c.show(show_ports=True)


if __name__ == "__main__":
    import gdsfactory as gf

    c = gf.Component("pads_bundle_steps")
    pt = c << gf.components.pad_array(
        gf.partial(gf.components.pad, size=(30, 30)),
        orientation=270,
        columns=3,
        spacing=(50, 0),
    )
    pb = c << gf.components.pad_array(orientation=90, columns=3)
    pt.move((300, 500))

    routes = get_bundle_from_steps_electrical(
        pb.ports, pt.ports, end_straight_length=60, separation=30, steps=[{"dy": 100}]
    )

    for route in routes:
        c.add(route.references)

    c.show(show_ports=True)
